"""Import service types."""

from typing import List, Union

from typing_extensions import TypedDict


class AssetLike(TypedDict, total=False):
    """General type of an asset object through the import functions."""

    content: Union[str, bytes, dict]
    multi_layer_content: Union[List[dict], None]
    json_content: Union[dict, str, list]
    external_id: str
    json_metadata: Union[str, dict]
    is_honeypot: bool
    id: str


class KiliResolverAsset(AssetLike, TypedDict, total=True):
    """Type of an asset object to be sent in Kili resolvers."""

    content: Union[str, bytes]
    multi_layer_content: List[dict]
    json_content: Union[dict, str, list]
    external_id: str
    json_metadata: Union[str, dict]
    is_honeypot: bool
    id: str



"""GraphQL payload data mappers for asset operations."""

from typing import Any, Dict

from kili.domain.issue import IssueFilters


def issue_where_mapper(filters: IssueFilters) -> Dict[str, Any]:
    """Build the GraphQL IssueWhere variable to be sent in an operation."""
    return {
        "project": {"id": filters.project_id},
        "asset": {"id": filters.asset_id},
        "assetIn": filters.asset_id_in,
        "status": filters.status,
        "type": filters.issue_type,
    }



"""Organization use cases."""

from typing import Dict, Generator

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.organization import (
    OrganizationFilters,
    OrganizationId,
    OrganizationMetricsFilters,
    OrganizationToCreateInput,
    OrganizationToUpdateInput,
)
from kili.domain.types import ListOrTuple
from kili.use_cases.base import BaseUseCases


class OrganizationUseCases(BaseUseCases):
    """Organization use cases."""

    def create_organization(self, organization_data: OrganizationToCreateInput) -> dict:
        """Create an organization."""
        return self._kili_api_gateway.create_organization(
            organization=organization_data,
        )

    def update_organization(
        self,
        organization_id: OrganizationId,
        organization_data: OrganizationToUpdateInput,
    ) -> dict:
        """Update an organization."""
        return self._kili_api_gateway.update_organization(
            organization_id=organization_id,
            organization_data=organization_data,
        )

    def list_organizations(
        self, where: OrganizationFilters, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """List organizations."""
        return self._kili_api_gateway.list_organizations(
            filters=where,
            fields=fields,
            description="List organizations",
            options=options,
        )

    def count_organizations(self, where: OrganizationFilters) -> int:
        """Count organizations."""
        return self._kili_api_gateway.count_organizations(filters=where)

    def get_organization_metrics(
        self, where: OrganizationMetricsFilters, fields: ListOrTuple[str]
    ) -> Dict:
        """Get organization metrics."""
        return self._kili_api_gateway.get_organization_metrics(filters=where, fields=fields)



"""Mixin extending Kili API Gateway class with Tags related operations."""

from typing import Dict, List

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import fragment_builder
from kili.domain.project import ProjectId
from kili.domain.tag import TagId
from kili.domain.types import ListOrTuple

from .operations import (
    GQL_CHECK_TAG,
    GQL_DELETE_TAG,
    GQL_UNCHECK_TAG,
    GQL_UPDATE_TAG,
    get_list_tags_by_org_query,
    get_list_tags_by_project_query,
)
from .types import UpdateTagReturnData


class TagOperationMixin(BaseOperationMixin):
    """GraphQL Mixin extending GraphQL Gateway class with Tags related operations."""

    def list_tags_by_org(self, fields: ListOrTuple[str]) -> List[Dict]:
        """Send a GraphQL request calling listTagsByOrg resolver."""
        fragment = fragment_builder(fields=fields)
        query = get_list_tags_by_org_query(fragment)
        result = self.graphql_client.execute(query)
        return result["data"]

    def list_tags_by_project(self, project_id: ProjectId, fields: ListOrTuple[str]) -> List[Dict]:
        """Send a GraphQL request calling listTagsByProject resolver."""
        fragment = fragment_builder(fields=fields)
        query = get_list_tags_by_project_query(fragment)
        variables = {"projectId": project_id}
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def check_tag(self, project_id: ProjectId, tag_id: TagId) -> TagId:
        """Send a GraphQL request calling checkTag resolver."""
        variables = {"data": {"tagId": tag_id, "projectId": project_id}}
        result = self.graphql_client.execute(GQL_CHECK_TAG, variables)
        return TagId(result["data"]["id"])

    def uncheck_tag(self, project_id: ProjectId, tag_id: TagId) -> TagId:
        """Send a GraphQL request calling uncheckTag resolver.

        WARNING: The resolved expects the tag to be checked for the project in the first place,
            do not call optimistically.
        """
        variables = {"projectId": project_id, "tagId": tag_id}
        result = self.graphql_client.execute(GQL_UNCHECK_TAG, variables)
        return TagId(result["data"]["id"])

    def update_tag(self, tag_id: TagId, label: str) -> UpdateTagReturnData:
        """Send a GraphQL request calling updateTag resolver."""
        variables = {"tagId": tag_id, "data": {"label": label}}
        result = self.graphql_client.execute(GQL_UPDATE_TAG, variables)
        return UpdateTagReturnData(
            affected_rows=result["data"]["affectedRows"],
            updated_tag_id=TagId(result["data"]["updatedTag"]["id"]),
        )

    def delete_tag(self, tag_id: TagId) -> bool:
        """Send a GraphQL request calling deleteTag resolver."""
        variables = {"tagId": tag_id}
        result = self.graphql_client.execute(GQL_DELETE_TAG, variables)
        return result["data"]



"""Organization client methods."""

from datetime import datetime
from typing import Dict, Generator, Iterable, List, Literal, Optional, overload

import pytz
from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.organization import (
    OrganizationFilters,
    OrganizationId,
    OrganizationMetricsFilters,
    OrganizationToCreateInput,
    OrganizationToUpdateInput,
)
from kili.domain.types import ListOrTuple
from kili.presentation.client.base import BaseClientMethods
from kili.use_cases.organization.use_cases import OrganizationUseCases
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class InternalOrganizationClientMethods(BaseClientMethods):
    """Organization client methods."""

    @typechecked
    def create_organization(
        self,
        name: str,
        address: str,
        zip_code: str,
        city: str,
        country: str,
    ) -> Dict:
        """Create an organization.

        WARNING: This method is for internal use only.

        Each user must be linked to an organization

        Args:
            name : Name of the organization
            address : Address of the organization
            zip_code : Zip code of the organization
            city : City of the organization
            country : Country of the organization

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        organization_use_case = OrganizationUseCases(self.kili_api_gateway)
        return organization_use_case.create_organization(
            OrganizationToCreateInput(
                name=name,
                address=address,
                zip_code=zip_code,
                city=city,
                country=country,
            ),
        )

    @typechecked
    def update_properties_in_organization(
        self,
        organization_id: str,
        name: Optional[str] = None,
        license: Optional[dict] = None,  # noqa: A002
    ) -> Dict:
        """Modify an organization.

        WARNING: This method is for internal use only.

        Args:
            organization_id: Identifier of the organization
            name: New name of the organization
            license: New license of the organization

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        use_cases = OrganizationUseCases(self.kili_api_gateway)
        return use_cases.update_organization(
            OrganizationId(organization_id),
            OrganizationToUpdateInput(name=name, license=license),
        )


@for_all_methods(log_call, exclude=["__init__"])
class OrganizationClientMethods(BaseClientMethods):
    """Organization client methods."""

    @overload
    def organizations(
        self,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("id", "name"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def organizations(
        self,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("id", "name"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def organizations(
        self,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("id", "name"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of organizations that match a set of criteria.

        Args:
            email: Email of a user of the organization
            organization_id: Identifier of the organization
            fields: All the fields to request among the possible fields for the organizations.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#organization) for all possible fields.
            first: Maximum number of organizations to return.
            skip: Number of skipped organizations (they are ordered by creation date)
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the organizations is returned.

        Returns:
            An iterable of organizations.

        Examples:
            >>> kili.organizations(organization_id=organization_id, fields=['users.email'])
            [{'users': [{'email': 'john@doe.com'}]}]
        """
        organization_use_cases = OrganizationUseCases(self.kili_api_gateway)
        organization_gen = organization_use_cases.list_organizations(
            OrganizationFilters(
                email=email,
                organization_id=OrganizationId(organization_id) if organization_id else None,
            ),
            fields,
            QueryOptions(disable_tqdm=disable_tqdm, first=first, skip=skip),
        )

        if as_generator:
            return organization_gen
        return list(organization_gen)

    @typechecked
    def count_organizations(
        self, email: Optional[str] = None, organization_id: Optional[str] = None
    ) -> int:
        """Count organizations that match a set of criteria.

        Args:
            email: Email of a user of the organization
            organization_id: Identifier of the organization

        Returns:
            An integer corresponding to the number of organizations that match the criteria.
        """
        where = OrganizationFilters(
            email=email,
            organization_id=OrganizationId(organization_id) if organization_id else None,
        )
        return OrganizationUseCases(self.kili_api_gateway).count_organizations(where)

    @typechecked
    def organization_metrics(
        self,
        organization_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        fields: ListOrTuple[str] = (
            "numberOfAnnotations",
            "numberOfHours",
            "numberOfLabeledAssets",
        ),
    ) -> Dict:
        """Get organization metrics.

        Args:
            organization_id: Identifier of the organization
            start_date: Start date of the metrics computation
            end_date: End date of the metrics computation
            fields: Fields to request for the organization metrics.

        Returns:
            A dictionary containing the metrics of the organization.
        """
        if start_date is None:
            start_date = datetime.now(tz=pytz.UTC)
        if end_date is None:
            end_date = datetime.now(tz=pytz.UTC)
        filters = OrganizationMetricsFilters(
            id=OrganizationId(organization_id), start_datetime=start_date, end_datetime=end_date
        )

        return OrganizationUseCases(self.kili_api_gateway).get_organization_metrics(filters, fields)



"""Label Importers."""

import csv
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Type

import yaml

from kili.core.helpers import get_file_paths_to_upload
from kili.domain.label import LabelType
from kili.domain.project import ProjectId
from kili.services.helpers import get_external_id_from_file_path
from kili.services.label_import.exceptions import (
    LabelParsingError,
    MissingMetadataError,
    MissingTargetJobError,
)
from kili.services.label_import.parser import (
    AbstractLabelParser,
    KiliRawLabelParser,
    YoloLabelParser,
)
from kili.services.label_import.types import Classes, LabelFormat
from kili.services.types import LogLevel
from kili.use_cases.asset.utils import AssetUseCasesUtils
from kili.use_cases.label import LabelUseCases
from kili.use_cases.label.types import LabelToCreateUseCaseInput

if TYPE_CHECKING:
    from kili.client import Kili


class LoggerParams(NamedTuple):
    """Contains all parameters related to logging."""

    disable_tqdm: Optional[bool]
    level: LogLevel


class AbstractLabelImporter(ABC):
    """Abstract Label Importer."""

    def __init__(
        self, kili: "Kili", logger_params: LoggerParams, input_format: LabelFormat
    ) -> None:
        self.kili = kili
        self.logger_params = logger_params
        self.input_format: LabelFormat = input_format

        self.logger = logging.getLogger("kili.services.label_import")
        self.logger.setLevel(logger_params.level)

    def process_from_files(  # pylint: disable=too-many-arguments
        self,
        labels_files: List[Path],
        meta_file_path: Optional[Path],
        project_id: ProjectId,
        target_job_name: Optional[str],
        model_name: Optional[str],
        is_prediction: bool,
        overwrite: bool,
    ):
        """Performs the import from the label files."""
        self._check_arguments_compatibility(meta_file_path, target_job_name)
        class_by_id = self._read_classes_from_meta_file(meta_file_path, self.input_format)
        label_parser_class = self._select_label_parser()
        label_parser = label_parser_class(class_by_id, target_job_name)

        self.logger.warning("Importing labels")
        labels = self.extract_from_files(labels_files, label_parser)
        label_type = "PREDICTION" if is_prediction else "DEFAULT"
        self.process_from_dict(
            project_id=project_id,
            labels=labels,
            label_type=label_type,
            overwrite=overwrite,
            model_name=model_name,
        )

        self.logger.warning("%d labels have been successfully imported", len(labels))

    def process_from_dict(  # pylint: disable=too-many-arguments
        self,
        project_id: Optional[str],
        labels: List[Dict],
        label_type: LabelType,
        overwrite: bool,
        model_name: Optional[str] = None,
    ) -> List:
        """Imports labels from a list of dictionaries representing labels."""
        should_retrieve_asset_ids = labels[0].get("asset_id") is None
        if should_retrieve_asset_ids:
            assert project_id
            asset_external_ids = [label["asset_external_id"] for label in labels]
            asset_id_map = AssetUseCasesUtils(
                self.kili.kili_api_gateway
            ).infer_ids_from_external_ids(asset_external_ids, ProjectId(project_id))
            labels = [
                {**label, "asset_id": asset_id_map[label["asset_external_id"]]} for label in labels
            ]

        labels_data = [
            LabelToCreateUseCaseInput(
                json_response=label["json_response"],
                asset_id=label.get("asset_id"),
                seconds_to_label=label.get("seconds_to_label"),
                model_name=model_name,
                author_id=label.get("author_id"),
                asset_external_id=None,
                label_type=label_type,
            )
            for label in labels
        ]

        return LabelUseCases(self.kili.kili_api_gateway).append_labels(
            disable_tqdm=self.logger_params.disable_tqdm,
            overwrite=overwrite,
            label_type=label_type,
            project_id=ProjectId(project_id) if project_id else None,
            fields=("id",),
            labels=labels_data,
        )

    @staticmethod
    @abstractmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        pass

    @classmethod
    @abstractmethod
    def _read_classes_from_meta_file(
        cls, meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Optional[Classes]:
        pass

    @abstractmethod
    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        pass

    @abstractmethod
    def _get_label_file_extension(self) -> str:
        pass

    def extract_from_files(
        self, labels_files: List[Path], label_parser: AbstractLabelParser
    ) -> List[Dict]:
        """Extracts the labels files and their metadata from the label files."""
        label_paths = get_file_paths_to_upload(
            [str(f) for f in labels_files],
            lambda path: path.endswith(self._get_label_file_extension()),
            verbose=self.logger_params.level in ["INFO", "DEBUG"],
        )
        if len(label_paths) == 0:
            raise ValueError(
                "No label files to upload. Check that the paths exist and file types are .json"
            )
        labels = []
        for path in label_paths:
            try:
                labels.append(
                    {
                        "json_response": label_parser.parse(Path(path)),
                        "asset_external_id": get_external_id_from_file_path(Path(path)),
                    }
                )
            except Exception as exc:
                raise LabelParsingError(f"Failed to parse the file {path}") from exc
        return labels


class YoloLabelImporter(AbstractLabelImporter):
    """Label importer in the yolo format."""

    def _get_label_file_extension(self) -> str:
        return ".txt"

    @staticmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        if meta_file_path is None:
            raise MissingMetadataError("Meta file is needed to import the label")
        if target_job_name is None:
            raise MissingTargetJobError("A target job name is needed to import the label")

    @staticmethod
    def _read_classes_from_meta_file(
        meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Classes:
        assert meta_file_path
        classes: Classes = Classes({})
        if input_format == "yolo_v4":
            try:
                # layout: id class\n
                with meta_file_path.open("r", encoding="utf-8") as m_f:
                    csv_reader = csv.reader(m_f, delimiter=" ")
                    classes = Classes({int(r[0]): r[1] for r in csv_reader if r[0] != " "})
            except (ValueError, IndexError):
                with meta_file_path.open("r", encoding="utf-8") as m_f:
                    classes = Classes(dict(enumerate(m_f.read().splitlines())))

        elif input_format == "yolo_v5":
            with meta_file_path.open("r", encoding="utf-8") as m_f:
                m_d = yaml.load(m_f, yaml.FullLoader)
                classes = Classes(m_d["names"])
        elif input_format == "yolo_v7":
            with meta_file_path.open("r", encoding="utf-8") as m_f:
                m_d = yaml.load(m_f, yaml.FullLoader)
                classes = Classes(dict(enumerate(m_d["names"])))
        else:
            raise NotImplementedError(f"The format f{input_format} does not have a metadata parser")

        return classes

    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        return YoloLabelParser


class KiliRawLabelImporter(AbstractLabelImporter):
    """# Label importer in the Kili format."""

    def _get_label_file_extension(self) -> str:
        return ".json"

    @staticmethod
    def _check_arguments_compatibility(
        meta_file_path: Optional[Path], target_job_name: Optional[str]
    ):
        pass

    @classmethod
    def _read_classes_from_meta_file(
        cls, meta_file_path: Optional[Path], input_format: LabelFormat
    ) -> Optional[Classes]:
        return None

    def _select_label_parser(self) -> Type[AbstractLabelParser]:
        return KiliRawLabelParser



"""Functions to import assets into a TEXT project."""

import json
import os
from enum import Enum
from typing import List, Optional, Tuple

from kili.core.helpers import is_url

from .base import (
    BaseAbstractAssetImporter,
    BatchParams,
    ContentBatchImporter,
)
from .exceptions import ImportValidationError
from .types import AssetLike


class LLMDataType(Enum):
    """LLM data type."""

    DICT = "DICT"
    LOCAL_FILE = "LOCAL_FILE"
    HOSTED_FILE = "HOSTED_FILE"


class JSONBatchImporter(ContentBatchImporter):
    """Class for importing a batch of LLM assets with dict content into a LLM_RLHF project."""

    def get_content_type_and_data_from_content(self, content: Optional[str]) -> Tuple[str, str]:
        """Returns the data of the content (path) and its content type."""
        return content or "", "application/json"


class LLMDataImporter(BaseAbstractAssetImporter):
    """Class for importing data into a TEXT project."""

    @staticmethod
    def get_data_type(assets: List[AssetLike]) -> LLMDataType:
        """Determine the type of data to upload from the service payload."""
        content_array = [asset.get("content", None) for asset in assets]
        if all(is_url(content) for content in content_array):
            return LLMDataType.HOSTED_FILE
        if all(isinstance(content, str) and os.path.exists(content) for content in content_array):
            return LLMDataType.LOCAL_FILE
        if all(isinstance(content, dict) for content in content_array):
            return LLMDataType.DICT
        raise ImportValidationError("Invalid value in content for LLM project.")

    def import_assets(self, assets: List[AssetLike]):
        """Import LLM assets into Kili."""
        self._check_upload_is_allowed(assets)
        data_type = self.get_data_type(assets)
        assets = self.filter_duplicate_external_ids(assets)
        if data_type == LLMDataType.LOCAL_FILE:
            assets = self.filter_local_assets(assets, self.raise_error)
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        elif data_type == LLMDataType.HOSTED_FILE:
            batch_params = BatchParams(is_hosted=True, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        elif data_type == LLMDataType.DICT:
            for asset in assets:
                if "content" in asset and isinstance(asset["content"], dict):
                    asset["content"] = json.dumps(asset["content"]).encode("utf-8")
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = JSONBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        else:
            raise ImportValidationError
        return self.import_assets_by_batch(assets, batch_importer)



"""Types for the Tag-related Kili API gateway functions."""

from dataclasses import dataclass

from kili.domain.tag import TagId


@dataclass
class UpdateTagReturnData:
    """UpdateTagReturn data.

    It is the updateTag resolver return data.
    """

    affected_rows: int
    updated_tag_id: TagId



"""Mixin extending Kili API Gateway class with Projects related operations."""

import json
from typing import Dict, Generator, Optional

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.adapters.kili_api_gateway.project.formatters import (
    PROJECT_JSON_FIELDS,
    load_project_json_fields,
)
from kili.adapters.kili_api_gateway.project.operations import get_projects_query
from kili.core.enums import ProjectType
from kili.domain.project import ComplianceTag, InputType, ProjectFilters, ProjectId
from kili.domain.types import ListOrTuple

from .common import get_project
from .mappers import project_data_mapper, project_where_mapper
from .operations import (
    GQL_COUNT_PROJECTS,
    GQL_CREATE_PROJECT,
    get_update_properties_in_project_mutation,
)
from .types import ProjectDataKiliAPIGatewayInput


class ProjectOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with Projects related operations."""

    def get_project(self, project_id: ProjectId, fields: ListOrTuple[str]) -> Dict:
        """Get project."""
        return get_project(self.graphql_client, project_id, fields)

    # pylint: disable=too-many-arguments
    def create_project(
        self,
        input_type: InputType,
        json_interface: Dict,
        title: str,
        description: str,
        project_type: Optional[ProjectType],
        compliance_tags: Optional[ListOrTuple[ComplianceTag]],
    ) -> ProjectId:
        """Create a project."""
        variables = {
            "data": {
                "description": description,
                "inputType": input_type,
                "jsonInterface": json.dumps(json_interface),
                "projectType": project_type,
                "title": title,
            }
        }
        # complience tags are only available for Kili app > 2.138
        if compliance_tags:
            variables["data"]["complianceTags"] = compliance_tags

        result = self.graphql_client.execute(GQL_CREATE_PROJECT, variables)
        return ProjectId(result["data"]["id"])

    def list_projects(
        self,
        project_filters: ProjectFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """Return a generator of projects that match the filter."""
        fragment = fragment_builder(fields)
        query = get_projects_query(fragment)
        where = project_where_mapper(filters=project_filters)
        projects_gen = PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving projects", GQL_COUNT_PROJECTS
        )
        if any(json_field in fields for json_field in PROJECT_JSON_FIELDS):
            projects_gen = (load_project_json_fields(project, fields) for project in projects_gen)
        return projects_gen

    def count_projects(self, project_filters: ProjectFilters) -> int:
        """Return the number of projects."""
        where = project_where_mapper(filters=project_filters)
        variables = {"where": where}
        result = self.graphql_client.execute(GQL_COUNT_PROJECTS, variables)
        return result["data"]

    def update_properties_in_project(
        self,
        project_id: ProjectId,
        project_data: ProjectDataKiliAPIGatewayInput,
        fields: ListOrTuple[str],
    ) -> Dict:
        """Update properties in a project."""
        fragment = fragment_builder(fields)
        mutation = get_update_properties_in_project_mutation(fragment)
        data = project_data_mapper(data=project_data)

        # complience tags are only available for Kili app > 2.138
        if "complianceTags" in data and data["complianceTags"] is None:
            del data["complianceTags"]

        variables = {"data": data, "where": {"id": project_id}}
        result = self.graphql_client.execute(mutation, variables)
        return load_project_json_fields(result["data"], fields)



"""Types for the Project-related Kili API gateway functions."""

from dataclasses import dataclass
from typing import Dict, Optional

from kili.domain.project import ComplianceTag, InputType
from kili.domain.types import ListOrTuple


@dataclass
# pylint: disable=too-many-instance-attributes
class ProjectDataKiliAPIGatewayInput:
    """Project input data for Kili API Gateway."""

    archived: Optional[bool]
    author: Optional[str]
    compliance_tags: Optional[ListOrTuple[ComplianceTag]]
    consensus_mark: Optional[float]
    consensus_tot_coverage: Optional[int]
    description: Optional[str]
    can_navigate_between_assets: Optional[bool]
    can_skip_asset: Optional[bool]
    honeypot_mark: Optional[float]
    input_type: Optional[InputType]
    instructions: Optional[str]
    json_interface: Optional[str]
    metadata_types: Optional[Dict]
    min_consensus_size: Optional[int]
    number_of_assets: Optional[int]
    rules: Optional[str]
    number_of_skipped_assets: Optional[int]
    number_of_remaining_assets: Optional[int]
    number_of_reviewed_assets: Optional[int]
    review_coverage: Optional[int]
    should_relaunch_kpi_computation: Optional[bool]
    title: Optional[str]
    use_honeypot: Optional[bool]



"""GraphQL Asset operations."""


def get_assets_query(fragment: str) -> str:
    """Return the GraphQL assets query."""
    return f"""
        query assets($where: AssetWhere!, $first: PageSize!, $skip: Int!) {{
            data: assets(where: $where, skip: $skip, first: $first) {{
                {fragment}
            }}
        }}
        """


GQL_COUNT_ASSETS = """
query countAssets($where: AssetWhere!) {
    data: countAssets(where: $where)
}
"""

GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS = """
query($filePaths: [String!]) {
  urls: createUploadBucketSignedUrls(filePaths: $filePaths)
}
"""



"""DEPRECATED.

From https://docs.kili-technology.com/reference/graphql-api#enums.
"""

from typing import Literal

LicenseType = Literal[
    "DISCOVERY",
    "FREEMIUM",
    "PAID",
]


LockType = Literal[
    "REVIEW",
    "DEFAULT",
]


NotificationStatus = Literal[
    "FAILURE",
    "PENDING",
    "SUCCESS",
]


OrganizationRole = Literal[
    "ADMIN",
    "USER",
]


ProjectRole = Literal[
    "ADMIN",
    "TEAM_MANAGER",
    "REVIEWER",
    "LABELER",
]


ProjectType = Literal[
    "IMAGE_CLASSIFICATION_SINGLE",
    "IMAGE_CLASSIFICATION_MULTI",
    "IMAGE_OBJECT_DETECTION_RECTANGLE",
    "IMAGE_OBJECT_DETECTION_POLYGON",
    "IMAGE_OBJECT_DETECTION_SEMANTIC",
    "IMAGE_POSE_ESTIMATION",
    "OCR",
    "PDF_CLASSIFICATION_SINGLE",
    "PDF_CLASSIFICATION_MULTI",
    "PDF_OBJECT_DETECTION_RECTANGLE",
    "PDF_NAMED_ENTITY_RECOGNITION",
    "SPEECH_TO_TEXT",
    "TEXT_CLASSIFICATION_SINGLE",
    "TEXT_CLASSIFICATION_MULTI",
    "TEXT_TRANSCRIPTION",
    "TEXT_NER",
    "TIME_SERIES",
    "VIDEO_CLASSIFICATION_SINGLE",
    "VIDEO_OBJECT_DETECTION",
    "VIDEO_FRAME_CLASSIFICATION",
    "VIDEO_FRAME_OBJECT_TRACKING",
]


Right = Literal[
    "CAN_ACCESS_SMART_TOOLS",
    "CAN_LABEL",
    "CREATE_AUDIO",
    "CREATE_FRAME",
    "CREATE_IMAGE",
    "CREATE_PDF",
    "CREATE_TEXT",
    "CREATE_TIMESERIES",
    "CREATE_VIDEO",
    "MAKE_PUBLIC_PROJECT",
    "SEE_LICENSE_BANNER",
    "SEE_UPGRADE_BUTTON",
    "UPLOAD_CLOUD_DATA",
    "USE_API",
    "USE_API_PRIORITY",
]



"""Module for common argument validators across client methods."""

import warnings
from typing import Any, List, Optional

from kili.domain.types import ListOrTuple


def disable_tqdm_if_as_generator(
    as_generator: bool, disable_tqdm: Optional[bool]
) -> Optional[bool]:
    """Disable tqdm in user-facing queries method if the return type is asked as a generator."""
    if as_generator and not disable_tqdm:
        disable_tqdm = True
        warnings.warn(
            "tqdm has been forced disabled because its behavior is not compatible with the"
            " generator return type",
            stacklevel=2,
        )
    return disable_tqdm


def assert_all_arrays_have_same_size(
    arrays: List[Optional[ListOrTuple[Any]]], raise_error: bool = True
) -> bool:
    """Assert that all given arrays have the same size if they are not None."""
    sizes_arrays = {len(array) for array in arrays if array is not None}
    if len(sizes_arrays) > 1:
        if raise_error:
            raise ValueError("All arrays should have the same length")
        return False
    return True



"""Functions to import assets into an IMAGE project."""

import os
from typing import List

from kili.core.constants import mime_extensions_that_need_post_processing
from kili.core.helpers import get_mime_type

from .base import BaseAbstractAssetImporter, BatchParams, ContentBatchImporter
from .constants import LARGE_IMAGE_THRESHOLD_SIZE
from .types import AssetLike


class ImageDataImporter(BaseAbstractAssetImporter):
    """Class for importing assets into an IMAGE project."""

    def import_assets(self, assets: List[AssetLike]):
        """Import IMAGE assets into Kili."""
        self._check_upload_is_allowed(assets)
        is_hosted = self.is_hosted_content(assets)
        if not is_hosted:
            assets = self.filter_local_assets(assets, self.raise_error)
        assets = self.filter_duplicate_external_ids(assets)
        sync_assets, async_assets = self.split_asset_by_upload_type(assets, is_hosted)
        created_asset_ids: List[str] = []
        if len(sync_assets) > 0:
            sync_batch_params = BatchParams(is_hosted=is_hosted, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, sync_batch_params, self.pbar
            )
            created_asset_ids += self.import_assets_by_batch(sync_assets, batch_importer)
        if len(async_assets) > 0:
            async_batch_params = BatchParams(is_hosted=is_hosted, is_asynchronous=True)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, async_batch_params, self.pbar
            )
            created_asset_ids += self.import_assets_by_batch(async_assets, batch_importer)
        return created_asset_ids

    @staticmethod
    def split_asset_by_upload_type(assets: List[AssetLike], is_hosted: bool):
        """Split assets into two groups, assets to to imported synchronously or asynchronously."""
        if is_hosted:
            return assets, []
        sync_assets, async_assets = [], []
        for asset in assets:
            multi_layer_content = asset.get("multi_layer_content")
            if multi_layer_content is not None:
                async_assets.append(asset)
                continue
            json_content = asset.get("json_content")
            path = asset.get("content")
            if json_content and not path:
                sync_assets.append(asset)
                continue
            assert path
            assert isinstance(path, str)
            mime_type = get_mime_type(path)
            is_large_image = os.path.getsize(path) >= LARGE_IMAGE_THRESHOLD_SIZE
            if is_large_image or mime_type in mime_extensions_that_need_post_processing:
                async_assets.append(asset)
            else:
                sync_assets.append(asset)
        return sync_assets, async_assets



"""GraphQL label operations."""


def get_labels_query(fragment: str) -> str:
    """Get the gql labels query."""
    return f"""
    query labels($where: LabelWhere!, $first: PageSize!, $skip: Int!) {{
        data: labels(where: $where, first: $first, skip: $skip) {{
            {fragment}
        }}
    }}
    """


GQL_COUNT_LABELS = """
query countLabels($where: LabelWhere!) {
    data: countLabels(where: $where)
}
"""


def get_update_properties_in_label_mutation(fragment: str) -> str:
    """Get the gql update properties in label query."""
    return f"""
    mutation(
        $where: LabelWhere!
        $data: LabelData!
    ) {{
    data: updatePropertiesInLabel(
        where: $where
        data:  $data
    ) {{
        {fragment}
    }}
    }}
    """


GQL_DELETE_LABELS = """
mutation($ids: [ID!]!) {
  data: deleteLabels(ids: $ids)
}
"""


def get_append_many_labels_mutation(fragment: str) -> str:
    """Get the gql append many labels mutation."""
    return f"""
    mutation appendManyLabels($data: AppendManyLabelsData!, $where: AssetWhere!) {{
        data: appendManyLabels(data: $data, where: $where ) {{
            {fragment}
        }}
    }}
    """


def get_create_honeypot_mutation(fragment: str) -> str:
    """Get the gql create honeypot mutation."""
    return f"""
    mutation(
        $data: CreateHoneypotData!
        $where: AssetWhere!
    ) {{
    data: createHoneypot(
        data: $data
        where: $where
    ) {{
        {fragment}
    }}
    }}
    """


def get_append_to_labels_mutation(fragment: str) -> str:
    """Get the gql append to labels mutation."""
    return f"""
mutation(
    $data: AppendToLabelsData!
    $where: AssetWhere!
) {{
  data: appendToLabels(
    data: $data
    where: $where
  ) {{
      {fragment}
  }}
}}
"""


def get_annotations_query(
    *,
    annotation_fragment: str,
    classification_annotation_fragment: str,
    ranking_annotation_fragment: str,
    transcription_annotation_fragment: str,
    video_annotation_fragment: str,
    video_object_detection_annotation_fragment: str,
    video_classification_annotation_fragment: str,
    video_transcription_annotation_fragment: str,
) -> str:
    """Get the gql annotations query."""
    inline_fragments = ""

    if classification_annotation_fragment.strip():
        inline_fragments += f"""
            ... on ClassificationAnnotation {{
                {classification_annotation_fragment}
            }}
        """

    if ranking_annotation_fragment.strip():
        inline_fragments += f"""
            ... on RankingAnnotation {{
                {ranking_annotation_fragment}
            }}
        """

    if transcription_annotation_fragment.strip():
        inline_fragments += f"""
            ... on TranscriptionAnnotation {{
                {transcription_annotation_fragment}
            }}
        """

    if video_annotation_fragment.strip():
        inline_fragments += f"""
            ... on VideoAnnotation {{
                    {video_annotation_fragment}
            }}
        """

    if video_object_detection_annotation_fragment.strip():
        inline_fragments += f"""
            ... on VideoObjectDetectionAnnotation {{
                {video_object_detection_annotation_fragment}
            }}
        """

    if video_classification_annotation_fragment.strip():
        inline_fragments += f"""
            ... on VideoClassificationAnnotation {{
                {video_classification_annotation_fragment}
            }}
        """

    if video_transcription_annotation_fragment.strip():
        inline_fragments += f"""
            ... on VideoTranscriptionAnnotation {{
                {video_transcription_annotation_fragment}
            }}
        """

    return f"""
    query annotations($where: AnnotationWhere!) {{
        data: annotations(where: $where) {{
            {annotation_fragment}
            {inline_fragments}
        }}
    }}
    """



"""GraphQL module."""

from typing import Any, Dict, Generator, NamedTuple, Optional

from typeguard import typechecked

from kili.core.constants import QUERY_BATCH_SIZE
from kili.core.graphql.graphql_client import GraphQLClient
from kili.domain.types import ListOrTuple
from kili.utils.tqdm import tqdm


class QueryOptions(NamedTuple):
    """Options when calling GraphQLQuery from the SDK."""

    disable_tqdm: Optional[bool]
    first: Optional[int] = None
    skip: int = 0


class PaginatedGraphQLQuery:
    """Query class for querying Kili objects.

    It factorizes code for executing paginated queries.
    """

    def __init__(self, graphql_client: GraphQLClient) -> None:
        """Initialize the paginator."""
        self._graphql_client = graphql_client

    # pylint: disable=too-many-arguments
    def execute_query_from_paginated_call(
        self,
        query: str,
        where: Dict[str, Any],
        options: QueryOptions,
        tqdm_desc: str,
        count_query: Optional[str],
    ) -> Generator[Dict, None, None]:
        """Build a row generator from paginated query calls with the first and skip pattern.

        Args:
            query: The object query to execute and to send to graphQL, in string format
            where: The where payload to send in the graphQL query
            options: The query options with skip and first and disable_tqdm
            tqdm_desc: The description to show in the progress bar
            count_query: The query to count the number of objects to be retrieved.
                It should have the same where input as the query.
                If given, it will show a progress bar if tqdm is not disabled in options
        """
        nb_elements_to_query = (
            self.get_number_of_elements_to_query(count_query, where, options)
            if count_query is not None
            else None
        )
        disable_tqdm = nb_elements_to_query is None or options.disable_tqdm

        if nb_elements_to_query == 0:
            yield from ()
        else:
            with tqdm(total=nb_elements_to_query, disable=disable_tqdm, desc=tqdm_desc) as pbar:
                count_elements_retrieved = 0
                while True:
                    if (
                        nb_elements_to_query is not None
                        and count_elements_retrieved >= nb_elements_to_query
                    ):
                        break

                    skip = count_elements_retrieved + options.skip
                    first = (
                        min(QUERY_BATCH_SIZE, nb_elements_to_query - count_elements_retrieved)
                        if nb_elements_to_query is not None
                        else QUERY_BATCH_SIZE
                    )
                    payload = {"where": where, "skip": skip, "first": first}
                    elements = self._graphql_client.execute(query, payload)["data"]
                    if not isinstance(elements, list):
                        raise TypeError(
                            "PaginatedGraphQLQuery only support operations returning a list of"
                            " objects"
                        )

                    if len(elements) == 0:
                        break

                    yield from elements

                    count_elements_retrieved += len(elements)
                    pbar.update(len(elements))

                    if len(elements) < first:
                        break

    def get_number_of_elements_to_query(
        self,
        count_query: str,
        where: Dict[str, Any],
        options: QueryOptions,
    ) -> int:
        """Give the total number of elements to query for one query that will be paginated.

        It uses both the argument first given by the user
        and the total number of available objects obtained with a graphQL count query

        Args:
            count_query: the query to count the number of objects to be retrieved
            where: the where payload to the count query,
            options: the query options with skip and first

        Returns:
            The number of elements to query
        """
        first = options.first
        skip = options.skip
        payload = {"where": where}
        count_result = self._graphql_client.execute(count_query, payload)
        nb_elements = count_result["data"]
        nb_elements_queried = max(nb_elements - skip, 0)
        if first is None:
            return nb_elements_queried
        return min(nb_elements_queried, first)


@typechecked
def fragment_builder(fields: ListOrTuple[str]) -> str:
    """Build a GraphQL fragment for a list of fields to query.

    Args:
        fields: The list of fields to query
    """
    fragment = ""

    # split a field and its subfields (e.g. "roles.user.id" -> ["roles", "user.id"])
    subfields = [field.split(".", 1) for field in fields if "." in field and "..." not in field]

    if subfields:
        # get the root fields (e.g. "roles" in "roles.user.id")
        root_fields = {subfield[0] for subfield in subfields}
        for root_field in root_fields:
            # get the subfields of the root field (e.g. "user.id" in "roles.user.id")
            fields_subquery = [subfield[1] for subfield in subfields if subfield[0] == root_field]
            # build the subquery fragment (e.g. "user{id}" in "roles{user{id}}")
            new_fragment = fragment_builder(fields_subquery)
            # add the subquery to the fragment
            fragment += f" {root_field}{{{new_fragment}}}"

        # remove the fields that have been queried in subqueries (e.g. "roles.user.id")
        fields = [field for field in fields if "." not in field]

    for field in fields:
        fragment += f" {field}"

    return fragment



"""Common code for the yolo exporter."""

import logging
from pathlib import Path
from typing import Dict, List, Set, Tuple

from kili.domain.ontology import JobMLTask, JobTool
from kili.services.export.exceptions import (
    NoCompatibleJobError,
    NotCompatibleInputType,
    NotCompatibleOptions,
)
from kili.services.export.format.base import AbstractExporter
from kili.services.export.media.video import cut_video
from kili.services.export.repository import AbstractContentRepository, DownloadError
from kili.services.export.types import JobCategory, LabelFormat, SplitOption
from kili.services.types import Job
from kili.utils.tqdm import tqdm


class YoloExporter(AbstractExporter):
    """Common code for Yolo exporters."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.split_option == "merged":
            self.merged_categories_id = self._get_merged_categories(self.project["jsonInterface"])
        else:
            self.categories_by_job = self._get_categories_by_job(self.project["jsonInterface"])

    def _check_arguments_compatibility(self) -> None:
        """Checks if the export label format is compatible with the export options."""
        if self.single_file:
            raise NotCompatibleOptions(
                f"The label format {self.label_format} can not be exported into a single file."
            )

        if self.normalized_coordinates is False:
            raise NotCompatibleOptions(
                "The YOLO annotation format does not support pixel coordinates."
            )

    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""
        if self.project["inputType"] not in ("IMAGE", "VIDEO"):
            raise NotCompatibleInputType(
                f"Project with input type '{self.project['inputType']}' not compatible with YOLO"
                " export format."
            )

        if len(self.compatible_jobs) == 0:
            raise NoCompatibleJobError("Project needs at least one compatible job.")

        if self.split_option == "merged":
            if not self.merged_categories_id:
                raise NoCompatibleJobError(
                    f"Error: There is no job in project {self.project_id} "
                    f"that can be converted to the {self.label_format} format."
                )
        else:
            if not self.categories_by_job:
                raise NoCompatibleJobError(
                    f"Error: There is no job in project {self.project_id} "
                    f"that can be converted to the {self.label_format} format."
                )

    def _is_job_compatible(self, job: Job) -> bool:
        """Check job compatibility with the YOLO format."""
        if "tools" not in job:
            return False

        compatible_tools = {JobTool.RECTANGLE, JobTool.POLYGON, JobTool.SEMANTIC}

        return job["mlTask"] == JobMLTask.OBJECT_DETECTION and all(
            tool in compatible_tools
            for tool in job["tools"]  # pyright: ignore[reportGeneralTypeIssues]
        )

    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """Yolo specific process and save."""
        if self.split_option == "merged":
            return self._process_and_save_merge(assets, output_filename)
        return self._process_and_save_split(assets, output_filename)

    def _process_and_save_split(self, assets: List[Dict], output_filename: Path) -> None:
        self.logger.info("Exporting to yolo format split...")

        self._write_jobs_labels_into_split_folders(
            assets,
            self.categories_by_job,
            self.export_root_folder,
            self.images_folder,
        )
        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def _process_and_save_merge(self, assets: List[Dict], output_filename: Path) -> None:
        self.logger.info("Exporting to yolo format merged...")

        labels_folder = self.base_folder / "labels"
        labels_folder.mkdir(parents=True, exist_ok=True)
        self._write_labels_into_single_folder(
            assets,
            self.merged_categories_id,
            labels_folder,
            self.images_folder,
            self.base_folder,
        )
        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def _write_labels_into_single_folder(
        self,
        assets: List[Dict],
        categories_id: Dict[str, JobCategory],
        labels_folder: Path,
        images_folder: Path,
        base_folder: Path,
    ):  # pylint: disable=too-many-arguments
        """Write all the labels into a single folder."""
        _write_class_file(base_folder, categories_id, self.label_format, self.split_option)

        remote_content = []
        video_metadata = {}

        for asset in tqdm(assets, disable=self.disable_tqdm):
            asset_remote_content, video_filenames = _process_asset(
                asset,
                images_folder,
                labels_folder,
                categories_id,
                self.content_repository,
                self.with_assets,
                self.project["inputType"],
            )
            if video_filenames:
                video_metadata[asset["externalId"]] = video_filenames
            remote_content.extend(asset_remote_content)

        if video_metadata:
            self.write_video_metadata_file(video_metadata, base_folder)

        if len(remote_content) > 0:
            self.images_folder.mkdir(parents=True, exist_ok=True)
            self.write_remote_content_file(remote_content, images_folder)

    def _write_jobs_labels_into_split_folders(
        self,
        assets: List[Dict],
        categories_by_job: Dict[str, Dict[str, JobCategory]],
        root_folder: Path,
        images_folder: Path,
    ) -> None:
        """Write assets into split folders."""
        for job_id, category_ids in categories_by_job.items():
            base_folder = root_folder / self.project_id / job_id
            labels_folder = base_folder / "labels"
            labels_folder.mkdir(parents=True, exist_ok=True)

            self._write_labels_into_single_folder(
                assets,
                category_ids,
                labels_folder,
                images_folder,
                base_folder,
            )

    def _get_merged_categories(self, json_interface: Dict) -> Dict[str, JobCategory]:
        """Return a dictionary of JobCategory instances by category full name."""
        cat_number = 0
        merged_categories_id: Dict[str, JobCategory] = {}
        for job_id, job in json_interface.get("jobs", {}).items():
            if not self._is_job_compatible(job):
                continue

            for category in job.get("content", {}).get("categories", {}):
                merged_categories_id[get_category_full_name(job_id, category)] = JobCategory(
                    category_name=category, id=cat_number, job_id=job_id
                )
                cat_number += 1

        return merged_categories_id

    def _get_categories_by_job(self, json_interface: Dict) -> Dict[str, Dict[str, JobCategory]]:
        """Return a dictionary of JobCategory instances by category full name and job id."""
        categories_by_job: Dict[str, Dict[str, JobCategory]] = {}
        for job_id, job in json_interface.get("jobs", {}).items():
            if not self._is_job_compatible(job):
                continue

            categories: Dict[str, JobCategory] = {}
            for cat_id, category in enumerate(job.get("content", {}).get("categories", {})):
                categories[get_category_full_name(job_id, category)] = JobCategory(
                    category_name=category, id=cat_id, job_id=job_id
                )
            categories_by_job[job_id] = categories
        return categories_by_job


class _LabelFrames:
    """Holds asset frames data."""

    @staticmethod
    def from_asset(asset, job_ids) -> "_LabelFrames":
        """Instantiate the label frames from the asset.

        It handles the case when there are several
        frames by label or a single one.
        """
        frames = {}
        number_of_frames = 0
        is_frame_group = False
        if "jsonResponse" in asset["latestLabel"]:
            number_of_frames = len(asset["latestLabel"]["jsonResponse"])
            for idx in range(number_of_frames):
                if str(idx) in asset["latestLabel"]["jsonResponse"]:
                    is_frame_group = True
                    frame_asset = asset["latestLabel"]["jsonResponse"][str(idx)]
                    for job_id in job_ids:
                        if (
                            job_id in frame_asset
                            and "annotations" in frame_asset[job_id]
                            and frame_asset[job_id]["annotations"]
                        ):
                            frames[idx] = {"latestLabel": {"jsonResponse": frame_asset}}
                            break

        if not frames:
            frames[-1] = asset
        return _LabelFrames(frames, number_of_frames, is_frame_group, asset["externalId"])

    def __init__(
        self, frames: Dict[int, Dict], number_frames: int, is_frame_group: bool, external_id: str
    ) -> None:
        self.frames: Dict[int, Dict] = frames
        self.number_frames: int = number_frames
        self.is_frame_group: bool = is_frame_group
        self.external_id: str = external_id

    def get_leading_zeros(self) -> int:
        """Get leading zeros for file name building."""
        return len(str(self.number_frames))

    def get_label_filename(self, idx: int) -> str:
        """Get label filemame for index."""
        return f"{self.external_id}_{str(idx + 1).zfill(self.get_leading_zeros())}"


def _convert_from_kili_to_yolo_format(
    job_id: str, label: Dict, category_ids: Dict[str, JobCategory]
) -> List[Tuple]:
    # pylint: disable=too-many-locals
    """Extract formatted annotations from labels and save the zip in the buckets."""
    if label is None or "jsonResponse" not in label:
        return []
    json_response = label["jsonResponse"]
    if not (job_id in json_response and "annotations" in json_response[job_id]):
        return []
    annotations = json_response[job_id]["annotations"]
    converted_annotations: List[Tuple] = []
    for annotation in annotations:
        category_idx: JobCategory = category_ids[
            get_category_full_name(job_id, annotation["categories"][0]["name"])
        ]
        if "boundingPoly" not in annotation:
            continue
        bounding_poly = annotation["boundingPoly"]
        if len(bounding_poly) < 1 or "normalizedVertices" not in bounding_poly[0]:
            continue
        normalized_vertices = bounding_poly[0]["normalizedVertices"]
        x_s: List[float] = [vertice["x"] for vertice in normalized_vertices]
        y_s: List[float] = [vertice["y"] for vertice in normalized_vertices]

        if annotation["type"] == JobTool.RECTANGLE:
            x_min, y_min = min(x_s), min(y_s)
            x_max, y_max = max(x_s), max(y_s)
            bbox_center_x, bbox_center_y = (x_min + x_max) / 2, (y_min + y_max) / 2  # type: ignore
            bbox_width, bbox_height = x_max - x_min, y_max - y_min  # type: ignore
            converted_annotations.append(
                (category_idx.id, bbox_center_x, bbox_center_y, bbox_width, bbox_height)
            )

        elif annotation["type"] in {JobTool.POLYGON, JobTool.SEMANTIC}:
            # <class-index> <x1> <y1> <x2> <y2> ... <xn> <yn>
            # Each segmentation label must have a minimum of 3 xy points (polygon)
            points = [val for pair in zip(x_s, y_s) for val in pair]
            converted_annotations.append((category_idx.id, *points))

    return converted_annotations


def get_category_full_name(job_id: str, category_name: str):
    """Return a full name to identify uniquely a category."""
    return f"{job_id}__{category_name}"


def _process_asset(
    asset: Dict,
    images_folder: Path,
    labels_folder: Path,
    category_ids: Dict[str, JobCategory],
    content_repository: AbstractContentRepository,
    with_assets: bool,
    project_input_type: str,
) -> Tuple[List[Tuple[str, str, str]], List[str]]:
    # pylint: disable=too-many-locals, too-many-arguments
    """Process an asset for all job_ids of category_ids."""
    asset_remote_content = []
    job_ids = {job_category.job_id for job_category in category_ids.values()}

    label_frames = _LabelFrames.from_asset(asset, job_ids)

    leading_zeros = 0
    if label_frames.is_frame_group:
        nbr_frames = label_frames.number_frames
        leading_zeros = len(str(nbr_frames))

    # If the asset is a video, we need to cut it into frames
    if project_input_type == "VIDEO" and asset["jsonContent"] == "" and with_assets:
        asset["jsonContent"] = cut_video(
            video_path=asset["content"],
            asset=asset,
            leading_zeros=leading_zeros,
            output_dir=images_folder,
        )

    content_frames = []
    if isinstance(asset["jsonContent"], list):
        content_frames = asset["jsonContent"]
        content_frames = [str(path) for path in content_frames]

    video_filenames = []

    for idx, frame in label_frames.frames.items():
        if label_frames.is_frame_group:
            filename = label_frames.get_label_filename(idx)
            video_filenames.append(filename)
        else:
            filename = asset["externalId"]

        frame_labels = _get_frame_labels(frame, job_ids, category_ids)

        _write_labels_to_file(labels_folder, filename, frame_labels)

        # no need to write asset urls since they are already downloaded
        if with_assets:
            continue

        content_frame = content_frames[idx] if content_frames else asset["content"]
        if content_repository.is_serving(content_frame):
            if content_frames and not label_frames.is_frame_group:
                try:
                    _write_content_frame_to_file(
                        content_frame, images_folder, filename, content_repository
                    )
                except DownloadError as download_error:
                    asset_id = asset["id"]
                    logging.warning("for asset %s: %s", asset_id, str(download_error))
        else:
            asset_remote_content.append([asset["externalId"], content_frame, f"{filename}.txt"])

    return asset_remote_content, video_filenames


def _write_class_file(
    folder: Path,
    category_ids: Dict[str, JobCategory],
    label_format: LabelFormat,
    layout: SplitOption,
):
    """Create a file that contains meta information about the export, depending of Yolo version.

    For the "merged" layout, the category name is prefixed with the job id to avoid duplicates
    since a same category name can be used in several jobs.
    """
    if label_format == "yolo_v4":
        with (folder / "classes.txt").open("wb") as fout:
            for job_category in category_ids.values():
                prefix = f"{job_category.job_id}/" if layout == "merged" else ""
                fout.write(f"{job_category.id} {prefix}{job_category.category_name}\n".encode())

    elif label_format == "yolo_v5":
        with (folder / "data.yaml").open("wb") as fout:
            fout.write(b"names:\n")
            for ind, job_category in enumerate(category_ids.values()):
                prefix = f"{job_category.job_id}/" if layout == "merged" else ""
                fout.write(f"  {ind}: {prefix}{job_category.category_name}\n".encode())

    elif label_format in ("yolo_v7", "yolo_v8"):
        with (folder / "data.yaml").open("wb") as fout:
            categories = ""
            for job_category in category_ids.values():
                prefix = f"{job_category.job_id}/" if layout == "merged" else ""
                categories += f"'{prefix}{job_category.category_name}', "
            fout.write(f"nc: {len(category_ids.items())}\n".encode())
            fout.write(f"names: [{categories[:-2]}]\n".encode())  # remove last comma

    else:
        raise ValueError(f"Unknown Yolo label format: {label_format}")


def _get_frame_labels(
    frame: Dict, job_ids: Set[str], category_ids: Dict[str, JobCategory]
) -> List[Tuple]:
    annotations = []
    for job_id in job_ids:
        job_annotations = _convert_from_kili_to_yolo_format(
            job_id, frame["latestLabel"], category_ids
        )
        annotations += job_annotations

    return annotations


def _write_content_frame_to_file(
    url_content_frame: str,
    images_folder: Path,
    filename: str,
    content_repository: AbstractContentRepository,
):
    content_iterator = content_repository.get_content_stream(url_content_frame, 1024)
    with (images_folder / f"{filename}.jpg").open("wb") as fout:
        for block in content_iterator:
            if not block:
                break
            fout.write(block)


def _write_labels_to_file(labels_folder: Path, filename: str, annotations: List[Tuple]) -> None:
    file_path = labels_folder / f"{filename}.txt"
    file_path.parent.mkdir(parents=True, exist_ok=True)
    with file_path.open("wb") as fout:
        for category_idx, *points in annotations:
            points_str = " ".join([str(point) for point in points])
            fout.write(f"{category_idx} {points_str}\n".encode())



"""Mixin extending Kili API Gateway class with Api Keys related operations."""

from datetime import datetime
from typing import Dict, Generator

from kili.adapters.kili_api_gateway.api_key.mappers import api_key_where_mapper
from kili.adapters.kili_api_gateway.api_key.operations import (
    GQL_API_KEY_EXPIRY_DATE,
    GQL_COUNT_API_KEYS,
    get_api_keys_query,
)
from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.domain.api_key import ApiKeyFilters
from kili.domain.types import ListOrTuple
from kili.exceptions import NotFound


class ApiKeyOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with API Keys related operations."""

    def list_api_keys(
        self,
        filters: ApiKeyFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List assets with given options."""
        fragment = fragment_builder(fields)
        query = get_api_keys_query(fragment)
        where = api_key_where_mapper(filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving api keys", GQL_COUNT_API_KEYS
        )

    def count_api_keys(self, filters: ApiKeyFilters) -> int:
        """Send a GraphQL request calling countIssues resolver."""
        where = api_key_where_mapper(filters)
        payload = {"where": where}
        count_result = self.graphql_client.execute(GQL_COUNT_API_KEYS, payload)
        count: int = count_result["data"]
        return count

    def get_api_key_expiry_date(self, api_key: str) -> datetime:
        """Get the expiry date of a given api key as a datetime type."""
        payload = {"where": {"key": api_key}}
        result = self.graphql_client.execute(GQL_API_KEY_EXPIRY_DATE, payload)
        fetched_key = next(iter(result["data"] or []), None)
        if fetched_key is None:
            raise NotFound(f"Could not find api key {api_key}")
        return datetime.strptime(fetched_key["expiryDate"], r"%Y-%m-%dT%H:%M:%S.%fZ")



"""GraphQL Project operations."""


def get_projects_query(fragment: str) -> str:
    """Return the GraphQL projects query."""
    return f"""
        query projects($where: ProjectWhere!, $first: PageSize!, $skip: Int!) {{
            data: projects(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """


def get_update_properties_in_project_mutation(fragment: str) -> str:
    """Return the GraphQL updatePropertiesInProject mutation."""
    return f"""
        mutation updatePropertiesInProject($data: ProjectData!, $where: ProjectWhere!) {{
            data: updatePropertiesInProject(data: $data, where: $where) {{
                {fragment}
            }}
        }}
        """


GQL_COUNT_PROJECTS = """
query countProjects($where: ProjectWhere!) {
    data: countProjects(where: $where)
}
"""

GQL_CREATE_PROJECT = """
mutation($data: CreateProjectData!) {
    data: createProject(data: $data) {
        id
    }
}
"""



"""Client presentation methods for labels."""

# pylint: disable=too-many-lines
import warnings
from itertools import repeat
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.helpers import (
    deprecate,
    is_empty_list_with_warning,
    validate_category_search_query,
)
from kili.domain.asset import AssetExternalId, AssetFilters, AssetId, AssetStatus
from kili.domain.asset.helpers import check_asset_identifier_arguments
from kili.domain.label import LabelFilters, LabelId, LabelType
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.domain.user import UserFilter, UserId
from kili.presentation.client.helpers.common_validators import (
    assert_all_arrays_have_same_size,
    disable_tqdm_if_as_generator,
)
from kili.services.export import export_labels
from kili.services.export.exceptions import NoCompatibleJobError
from kili.services.export.types import CocoAnnotationModifier, LabelFormat, SplitOption
from kili.use_cases.asset.utils import AssetUseCasesUtils
from kili.use_cases.label import LabelUseCases
from kili.use_cases.label.types import LabelToCreateUseCaseInput
from kili.utils.labels.parsing import ParsedLabel
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods

if TYPE_CHECKING:
    import pandas as pd


@for_all_methods(log_call, exclude=["__init__"])
class LabelClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on labels."""

    @typechecked
    def count_labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        asset_external_id_strictly_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        label_id: Optional[str] = None,
        type_in: Optional[List[LabelType]] = None,
        user_id: Optional[str] = None,
        category_search: Optional[str] = None,
        id_contains: Optional[List[str]] = None,
    ) -> int:
        # pylint: disable=line-too-long
        """Get the number of labels for the given parameters.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            asset_external_id_strictly_in: Returned labels should have an external id that exactly matches one of the ids in that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            label_id: Identifier of the label.
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            category_search: Query to filter labels based on the content of their jsonResponse
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of labels with the parameters provided
        """
        if category_search:
            validate_category_search_query(category_search)

        filters = LabelFilters(
            project_id=ProjectId(project_id),
            asset=AssetFilters(
                project_id=ProjectId(project_id),
                asset_id=AssetId(asset_id) if asset_id else None,
                status_in=asset_status_in,
                external_id_in=(
                    cast(ListOrTuple[AssetExternalId], asset_external_id_in)
                    if asset_external_id_in
                    else None
                ),
                external_id_strictly_in=(
                    cast(ListOrTuple[AssetExternalId], asset_external_id_strictly_in)
                    if asset_external_id_strictly_in
                    else None
                ),
            ),
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_in=cast(List[LabelId], id_contains) if id_contains else None,
            id=LabelId(label_id) if label_id else None,
            type_in=type_in,
            user=UserFilter(id=UserId(user_id)) if user_id else None,
            search=category_search,
            consensus_mark_gte=None,
            consensus_mark_lte=None,
            labeler_in=None,
            reviewer_in=None,
        )

        return LabelUseCases(self.kili_api_gateway).count_labels(filters=filters)

    @overload
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        asset_external_id_strictly_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
            "isLatestLabelForUser",
            "assetId",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[LabelType]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        output_format: Literal["dict"] = "dict",
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        asset_external_id_strictly_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
            "isLatestLabelForUser",
            "assetId",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[LabelType]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        output_format: Literal["dict"] = "dict",
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @overload
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        asset_external_id_strictly_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
            "isLatestLabelForUser",
            "assetId",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[LabelType]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        output_format: Literal["parsed_label"] = "parsed_label",
        *,
        as_generator: Literal[False] = False,
    ) -> List[ParsedLabel]:
        ...

    @overload
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        asset_external_id_strictly_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
            "isLatestLabelForUser",
            "assetId",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[LabelType]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        output_format: Literal["parsed_label"] = "parsed_label",
        *,
        as_generator: Literal[True] = True,
    ) -> Generator[ParsedLabel, None, None]:
        ...

    @typechecked
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        asset_external_id_strictly_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
            "isLatestLabelForUser",
            "assetId",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[LabelType]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        output_format: Literal["dict", "parsed_label"] = "dict",
        *,
        as_generator: bool = False,
    ) -> Iterable[Union[Dict, ParsedLabel]]:
        # pylint: disable=line-too-long
        """Get a label list or a label generator from a project based on a set of criteria.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED`, `TO REVIEW` or `REVIEWED`.
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            asset_external_id_strictly_in: Returned labels should have an external id that exactly matches one of the ids in that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have their creation date equal to this date.
            created_at_gte: Returned labels should have their creation date greater or equal to this date.
            created_at_lte: Returned labels should have their creation date lower or equal to this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the labels is returned.
            category_search: Query to filter labels based on the content of their jsonResponse.
            output_format: If `dict`, the output is an iterable of Python dictionaries.
                If `parsed_label`, the output is an iterable of parsed labels objects. More information on parsed labels in the [documentation](https://python-sdk-docs.kili-technology.com/latest/sdk/tutorials/label_parsing/).

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            An iterable of labels.

        Examples:
            >>> kili.labels(project_id=project_id, fields=['jsonResponse', 'labelOf.externalId']) # returns a list of all labels of a project and their assets external ID
            >>> kili.labels(project_id=project_id, fields=['jsonResponse'], as_generator=True) # returns a generator of all labels of a project

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators

        Example:
                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                category_search = `(JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
        """
        if category_search:
            validate_category_search_query(category_search)

        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)

        filters = LabelFilters(
            project_id=ProjectId(project_id),
            asset=AssetFilters(
                project_id=ProjectId(project_id),
                asset_id=AssetId(asset_id) if asset_id else None,
                status_in=asset_status_in,
                external_id_in=(
                    cast(ListOrTuple[AssetExternalId], asset_external_id_in)
                    if asset_external_id_in
                    else None
                ),
                external_id_strictly_in=(
                    cast(ListOrTuple[AssetExternalId], asset_external_id_strictly_in)
                    if asset_external_id_strictly_in
                    else None
                ),
            ),
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_in=cast(List[LabelId], id_contains) if id_contains else None,
            id=LabelId(label_id) if label_id else None,
            type_in=type_in,
            user=UserFilter(id=UserId(user_id)) if user_id else None,
            search=category_search,
            consensus_mark_gte=None,
            consensus_mark_lte=None,
            labeler_in=None,
            reviewer_in=None,
        )

        labels_gen = LabelUseCases(self.kili_api_gateway).list_labels(
            filters=filters,
            options=options,
            fields=fields,
            output_format=output_format,
            project_id=ProjectId(project_id),
        )

        if as_generator:
            return labels_gen
        return list(labels_gen)

    @overload
    def predictions(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "modelName",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def predictions(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "modelName",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def predictions(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "modelName",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get prediction labels from a project based on a set of criteria.

        This method is equivalent to the `kili.labels()` method, but it only returns label of type "PREDICTION".

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED`, `TO REVIEW` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.
            category_search: Query to filter labels based on the content of their jsonResponse

        Returns:
            An iterable of labels.

        Examples:
            >>> kili.predictions(project_id=project_id) # returns a list of prediction labels of a project
        """
        return self.labels(
            project_id=project_id,
            asset_id=asset_id,
            asset_status_in=asset_status_in,
            asset_external_id_in=asset_external_id_in,
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            fields=fields,
            first=first,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_contains=id_contains,
            label_id=label_id,
            skip=skip,
            type_in=["PREDICTION"],
            user_id=user_id,
            disable_tqdm=disable_tqdm,
            category_search=category_search,
            as_generator=as_generator,  # pyright: ignore[reportGeneralTypeIssues]
        )

    @overload
    def inferences(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "modelName",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def inferences(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "modelName",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def inferences(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[ListOrTuple[AssetStatus]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "modelName",
        ),
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: Optional[bool] = None,
        category_search: Optional[str] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get inference labels from a project based on a set of criteria.

        This method is equivalent to the `kili.labels()` method, but it only returns label of type "INFERENCE".

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED`, `TO REVIEW` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.
            category_search: Query to filter labels based on the content of their jsonResponse

        Returns:
            An iterable of inference labels.

        Examples:
            >>> kili.inferences(project_id=project_id) # returns a list of inference labels of a project
        """
        return self.labels(
            project_id=project_id,
            asset_id=asset_id,
            asset_status_in=asset_status_in,
            asset_external_id_in=asset_external_id_in,
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            fields=fields,
            first=first,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_contains=id_contains,
            label_id=label_id,
            skip=skip,
            type_in=["INFERENCE"],
            user_id=user_id,
            disable_tqdm=disable_tqdm,
            category_search=category_search,
            as_generator=as_generator,  # pyright: ignore[reportGeneralTypeIssues]
        )

    @typechecked
    def update_properties_in_label(
        self,
        label_id: str,
        seconds_to_label: Optional[int] = None,
        model_name: Optional[str] = None,
        json_response: Optional[dict] = None,
    ) -> Dict[Literal["id"], str]:
        """Update properties of a label.

        Args:
            label_id: Identifier of the label
            seconds_to_label: Time to create the label
            model_name: Name of the model
            json_response: The label is given here

        Returns:
            A dictionary with the label `id`.

        Examples:
            >>> kili.update_properties_in_label(label_id=label_id, json_response={...})
        """
        return LabelUseCases(self.kili_api_gateway).update_properties_in_label(
            label_id=LabelId(label_id),
            seconds_to_label=seconds_to_label,
            model_name=model_name,
            json_response=json_response,
            fields=("id",),
        )

    @typechecked
    def delete_labels(
        self,
        ids: ListOrTuple[str],
        disable_tqdm: Optional[bool] = None,
    ) -> List[str]:
        """Delete labels.

        Currently, only `PREDICTION` and `INFERENCE` labels can be deleted.

        Args:
            ids: List of label ids to delete.
            disable_tqdm: If `True`, the progress bar will be disabled.

        Returns:
            The deleted label ids.
        """
        if is_empty_list_with_warning("delete_labels", "ids", ids):
            return []

        deleted_label_ids = LabelUseCases(self.kili_api_gateway).delete_labels(
            ids=cast(ListOrTuple[LabelId], ids),
            disable_tqdm=disable_tqdm,
        )
        return cast(List[str], deleted_label_ids)

    @typechecked
    def append_labels(
        self,
        asset_id_array: Optional[List[str]] = None,
        json_response_array: ListOrTuple[Dict] = (),
        author_id_array: Optional[List[str]] = None,
        seconds_to_label_array: Optional[List[int]] = None,
        model_name: Optional[str] = None,
        label_type: LabelType = "DEFAULT",
        project_id: Optional[str] = None,
        asset_external_id_array: Optional[List[str]] = None,
        disable_tqdm: Optional[bool] = None,
        overwrite: bool = False,
    ) -> List[Dict[Literal["id"], str]]:
        """Append labels to assets.

        Args:
            asset_id_array: list of asset internal ids to append labels on.
            json_response_array: list of labels to append.
            author_id_array: list of the author id of the labels.
            seconds_to_label_array: list of times taken to produce the label, in seconds.
            model_name: Name of the model that generated the labels.
                Only useful when uploading PREDICTION or INFERENCE labels.
            label_type: Can be one of `AUTOSAVE`, `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`.
            project_id: Identifier of the project.
            asset_external_id_array: list of asset external ids to append labels on.
            disable_tqdm: Disable tqdm progress bar.
            overwrite: when uploading prediction or inference labels, if True,
                it will overwrite existing labels with the same model name
                and of the same label type, on the targeted assets.

        Returns:
            A list of dictionaries with the label ids.

        Examples:
            >>> kili.append_labels(
                    asset_id_array=['cl9wmlkuc00050qsz6ut39g8h', 'cl9wmlkuw00080qsz2kqh8aiy'],
                    json_response_array=[{...}, {...}]
                )
        """
        if len(json_response_array) == 0:
            raise ValueError(
                "json_response_array is empty, you must provide at least one label to upload"
            )

        check_asset_identifier_arguments(
            ProjectId(project_id) if project_id else None,
            cast(ListOrTuple[AssetId], asset_id_array) if asset_id_array else None,
            (
                cast(ListOrTuple[AssetExternalId], asset_external_id_array)
                if asset_external_id_array
                else None
            ),
        )

        assert_all_arrays_have_same_size(
            [
                seconds_to_label_array,
                author_id_array,
                json_response_array,
                asset_external_id_array,
                asset_id_array,
            ]
        )

        labels = [
            LabelToCreateUseCaseInput(
                asset_id=AssetId(asset_id) if asset_id else None,
                asset_external_id=AssetExternalId(asset_external_id) if asset_external_id else None,
                json_response=json_response,
                seconds_to_label=seconds_to_label,
                author_id=UserId(author_id) if author_id else None,
                label_type=label_type,
                model_name=model_name,
            )
            for (asset_id, asset_external_id, json_response, seconds_to_label, author_id) in zip(
                asset_id_array or repeat(None),
                asset_external_id_array or repeat(None),
                json_response_array,
                seconds_to_label_array or repeat(None),
                author_id_array or repeat(None),
            )
        ]

        return LabelUseCases(self.kili_api_gateway).append_labels(
            fields=("id",),
            disable_tqdm=disable_tqdm,
            label_type=label_type,
            labels=labels,
            overwrite=overwrite,
            project_id=ProjectId(project_id) if project_id else None,
        )

    @typechecked
    def create_predictions(
        self,
        project_id: str,
        external_id_array: Optional[List[str]] = None,
        model_name_array: Optional[List[str]] = None,
        json_response_array: Optional[List[dict]] = None,
        model_name: Optional[str] = None,
        asset_id_array: Optional[List[str]] = None,
        disable_tqdm: Optional[bool] = None,
        overwrite: bool = False,
    ) -> Dict[Literal["id"], str]:
        # pylint: disable=line-too-long
        """Create predictions for specific assets.

        Args:
            project_id: Identifier of the project.
            external_id_array: The external IDs of the assets for which we want to add predictions.
            model_name_array: Deprecated, use `model_name` instead.
            json_response_array: The predictions are given here. For examples,
                see [the recipe](https://docs.kili-technology.com/recipes/importing-labels-and-predictions).
            model_name: The name of the model that generated the predictions
            asset_id_array: The internal IDs of the assets for which we want to add predictions.
            disable_tqdm: Disable tqdm progress bar.
            overwrite: if True, it will overwrite existing predictions of
                the same model name on the targeted assets.

        Returns:
            A dictionary with the project `id`.

        !!! example "Recipe"
            For more detailed examples on how to create predictions, see [the recipe](https://docs.kili-technology.com/recipes/importing-labels-and-predictions).

        !!! warning "model name"
            The use of `model_name_array` is deprecated. Creating predictions from different
            models is not supported anymore. Please use `model_name` argument instead to
            provide the predictions model name.
        """
        if json_response_array is None or len(json_response_array) == 0:
            raise ValueError(
                "json_response_array is empty, you must provide at least one prediction to upload"
            )

        assert_all_arrays_have_same_size(
            [external_id_array, json_response_array, model_name_array, asset_id_array]
        )
        nb_labels_to_add = len(json_response_array)

        if model_name is None:
            if model_name_array is None:
                raise ValueError("You must provide a model name with the `model_name` argument.")

            if len(set(model_name_array)) > 1:
                raise ValueError(
                    "Creating predictions from different models is not supported anymore. Separate"
                    " your calls by models."
                )

            warnings.warn(
                "The use of `model_name_array` is deprecated. Creating predictions from"
                " different models is not supported anymore. Please use `model_name` argument"
                " instead to provide the predictions model name.",
                DeprecationWarning,
                stacklevel=1,
            )
            model_name = model_name_array[0]

        labels = [
            LabelToCreateUseCaseInput(
                asset_id=AssetId(asset_id) if asset_id else None,
                asset_external_id=AssetExternalId(asset_external_id) if asset_external_id else None,
                json_response=json_response,
                label_type="PREDICTION",
                model_name=model_name,
                seconds_to_label=None,
                author_id=None,
            )
            for (asset_id, asset_external_id, json_response) in zip(
                asset_id_array or repeat(None, nb_labels_to_add),
                external_id_array or repeat(None, nb_labels_to_add),
                json_response_array,
            )
        ]

        LabelUseCases(self.kili_api_gateway).append_labels(
            fields=("id",),
            disable_tqdm=disable_tqdm,
            label_type="PREDICTION",
            labels=labels,
            overwrite=overwrite,
            project_id=ProjectId(project_id) if project_id else None,
        )
        return {"id": project_id}

    @typechecked
    def create_honeypot(
        self,
        json_response: dict,
        asset_external_id: Optional[str] = None,
        asset_id: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> Dict:
        """Create honeypot for an asset.

        !!! info
            Uses the given `json_response` to create a `REVIEW` label.
            This enables Kili to compute a`honeypotMark`,
            which measures the similarity between this label and other labels.

        Args:
            json_response: The JSON response of the honeypot label of the asset.
            asset_id: Identifier of the asset.
                Either provide `asset_id` or `asset_external_id` and `project_id`.
            asset_external_id: External identifier of the asset.
                Either provide `asset_id` or `asset_external_id` and `project_id`.
            project_id: Identifier of the project.
                Either provide `asset_id` or `asset_external_id` and `project_id`.

        Returns:
            A dictionary-like object representing the created label.
        """
        return LabelUseCases(self.kili_api_gateway).create_honeypot_label(
            json_response=json_response,
            asset_id=AssetId(asset_id) if asset_id else None,
            asset_external_id=AssetExternalId(asset_external_id) if asset_external_id else None,
            project_id=ProjectId(project_id) if project_id else None,
            fields=("id",),
        )

    @deprecate(
        msg=(
            "append_to_labels method is deprecated. Please use append_labels instead. This new"
            " function allows to import several labels 10 times faster."
        )
    )
    @typechecked
    def append_to_labels(
        self,
        json_response: dict,
        author_id: Optional[str] = None,
        label_asset_external_id: Optional[str] = None,
        label_asset_id: Optional[str] = None,
        label_type: LabelType = "DEFAULT",
        project_id: Optional[str] = None,
        seconds_to_label: Optional[int] = 0,
    ) -> Dict[Literal["id"], str]:
        """!!! danger "[DEPRECATED]".

        append_to_labels method is deprecated. Please use append_labels instead.
            This new function allows to import several labels 10 times faster.

        Append a label to an asset.

        Args:
            json_response: Label is given here.
            author_id: ID of the author of the label.
            label_asset_external_id: External identifier of the asset.
            label_asset_id: Identifier of the asset.
            project_id: Identifier of the project.
            label_type: Can be one of `AUTOSAVE`, `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`.
            seconds_to_label: Time to create the label.

        !!! warning
            Either provide `label_asset_id` or `label_asset_external_id` and `project_id`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(label_asset_id=asset_id, json_response={...})
        """
        check_asset_identifier_arguments(
            ProjectId(project_id) if project_id else None,
            cast(ListOrTuple[AssetId], [label_asset_id]) if label_asset_id else None,
            (
                cast(ListOrTuple[AssetExternalId], [label_asset_external_id])
                if label_asset_external_id
                else None
            ),
        )

        if (
            label_asset_id is None
            and label_asset_external_id is not None
            and project_id is not None
        ):
            label_asset_id = AssetUseCasesUtils(self.kili_api_gateway).infer_ids_from_external_ids(
                cast(List[AssetExternalId], [label_asset_external_id]), ProjectId(project_id)
            )[AssetExternalId(label_asset_external_id)]

        return LabelUseCases(self.kili_api_gateway).append_to_labels(
            author_id=UserId(author_id) if author_id else None,
            json_response=json_response,
            label_type=label_type,
            asset_id=AssetId(label_asset_id),  # pyright: ignore[reportGeneralTypeIssues]
            seconds_to_label=seconds_to_label,
            fields=("id",),
        )

    @typechecked
    def export_labels_as_df(
        self,
        project_id: str,
        fields: ListOrTuple[str] = ("author.email", "author.id", "createdAt", "id", "labelType"),
        asset_fields: ListOrTuple[str] = ("externalId",),
    ) -> "pd.DataFrame":
        # pylint: disable=line-too-long
        """Get the labels of a project as a pandas DataFrame.

        Args:
            project_id: Identifier of the project
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            asset_fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.

        Returns:
            A pandas DataFrame containing the labels.
        """
        return LabelUseCases(self.kili_api_gateway).export_labels_as_df(
            project_id=ProjectId(project_id), label_fields=fields, asset_fields=asset_fields
        )

    def export_labels(
        self,
        project_id: str,
        filename: str,
        fmt: LabelFormat,
        asset_ids: Optional[List[str]] = None,
        layout: SplitOption = "split",
        single_file: bool = False,
        disable_tqdm: Optional[bool] = None,
        with_assets: bool = True,
        external_ids: Optional[List[str]] = None,
        annotation_modifier: Optional[CocoAnnotationModifier] = None,
        asset_filter_kwargs: Optional[Dict[str, object]] = None,
        normalized_coordinates: Optional[bool] = None,
    ) -> None:
        # pylint: disable=line-too-long
        """Export the project labels with the requested format into the requested output path.

        Args:
            project_id: Identifier of the project.
            filename: Relative or full path of the archive that will contain
                the exported data.
            fmt: Format of the exported labels.
            asset_ids: Optional list of the assets internal IDs from which to export the labels.
            layout: Layout of the exported files. "split" means there is one folder
                per job, "merged" that there is one folder with every labels.
            single_file: Layout of the exported labels. Single file mode is
                only available for some specific formats (COCO and Kili).
            disable_tqdm: Disable the progress bar if True.
            with_assets: Download the assets in the export.
            external_ids: Optional list of the assets external IDs from which to export the labels.
            annotation_modifier: (For COCO export only) function that takes the COCO annotation, the
                COCO image, and the Kili annotation, and should return an updated COCO annotation.
                This can be used if you want to add a new attribute to the COCO annotation. For
                example, you can add a method that computes if the annotation is a rectangle or not
                and add it to the COCO annotation (see example).
            asset_filter_kwargs: Optional dictionary of arguments to pass to `kili.assets()` in order to filter the assets the labels are exported from. The supported arguments are:

                - `consensus_mark_gte`
                - `consensus_mark_lte`
                - `external_id_strictly_in`
                - `external_id_in`
                - `honeypot_mark_gte`
                - `honeypot_mark_lte`
                - `label_author_in`
                - `label_reviewer_in`
                - `skipped`
                - `status_in`
                - `label_category_search`
                - `created_at_gte`
                - `created_at_lte`
                - `issue_type`
                - `issue_status`
                - `inference_mark_gte`
                - `inference_mark_lte`
                - `metadata_where`

                See the documentation of [`kili.assets()`](https://python-sdk-docs.kili-technology.com/latest/sdk/asset/#kili.queries.asset.__init__.QueriesAsset.assets) for more information.
            normalized_coordinates: This parameter is only effective on the Kili (a.k.a raw) format.
                If True, the coordinates of the `(x, y)` vertices are normalized between 0 and 1.
                If False, the json response will contain additional fields with coordinates in absolute values, that is, in pixels.

        !!! Info
            The supported formats are:

            - Yolo V4, V5, V7, V8 for object detection tasks.
            - Kili (a.k.a raw) for all tasks.
            - COCO for object detection tasks (bounding box and semantic segmentation).
            - Pascal VOC for object detection tasks (bounding box).

        !!! warning "Cloud storage"
            Export with asset download (`with_assets=True`) is not allowed for projects connected to a cloud storage.

        !!! Example
            ```python
            kili.export_labels("your_project_id", "export.zip", "yolo_v4")
            ```

        !!! Example
            ```python
            def is_rectangle(coco_annotation, coco_image, kili_annotation):
                is_rectangle = ...
                return {**coco_annotation, "attributes": {"is_rectangle": is_rectangle}}

            kili.export_labels(
                "your_project_id",
                "export.zip",
                "coco",
                annotation_modifier=add_is_rectangle
            )
            ```
        """
        if external_ids is not None and asset_ids is None:
            id_map = AssetUseCasesUtils(self.kili_api_gateway).infer_ids_from_external_ids(
                asset_external_ids=cast(List[AssetExternalId], external_ids),
                project_id=ProjectId(project_id),
            )
            resolved_asset_ids = [id_map[AssetExternalId(i)] for i in external_ids]
        else:
            resolved_asset_ids = cast(List[AssetId], asset_ids)

        try:
            export_labels(
                self,  # pyright: ignore[reportGeneralTypeIssues]
                asset_ids=resolved_asset_ids,
                project_id=ProjectId(project_id),
                export_type="latest",
                label_format=fmt,
                split_option=layout,
                single_file=single_file,
                output_file=filename,
                disable_tqdm=disable_tqdm,
                log_level="WARNING",
                with_assets=with_assets,
                annotation_modifier=annotation_modifier,
                asset_filter_kwargs=asset_filter_kwargs,
                normalized_coordinates=normalized_coordinates,
            )
        except NoCompatibleJobError as excp:
            warnings.warn(str(excp), stacklevel=2)



"""Pagination utils."""

from itertools import islice
from time import sleep
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, TypeVar

from kili.core.constants import MUTATION_BATCH_SIZE
from kili.domain.types import ListOrTuple
from kili.exceptions import GraphQLError


def batch_object_builder(
    properties_to_batch: Dict[str, ListOrTuple[Any]],
    batch_size: int = MUTATION_BATCH_SIZE,
) -> Generator[Dict[str, Any], None, None]:
    """Generate a paginated iterator for several variables.

    Args:
        properties_to_batch: a dictionary of properties to be batched.
        batch_size: the size of the batches to produce
    """
    if len(list(filter(None, properties_to_batch.values()))) == 0:
        yield properties_to_batch
        return
    # pylint: disable=stop-iteration-return
    number_of_objects = len(next(v for v in properties_to_batch.values() if v is not None))
    number_of_batches = len(range(0, number_of_objects, batch_size))
    batched_properties = {
        k: (
            batcher(iterable=v, batch_size=batch_size)
            if v is not None
            else (item for item in [v] * number_of_batches)
        )
        for k, v in properties_to_batch.items()
    }
    batch_object_iterator = (
        dict(zip(batched_properties, t)) for t in zip(*batched_properties.values())
    )
    yield from batch_object_iterator


# pylint: disable=missing-type-doc
def mutate_from_paginated_call(
    kili,
    properties_to_batch: Dict[str, ListOrTuple[Any]],
    generate_variables: Callable,
    request: str,
    batch_size: int = MUTATION_BATCH_SIZE,
    last_batch_callback: Optional[Callable] = None,
) -> List:
    """Run a mutation by making paginated calls.

    Args:
        kili: kili
        properties_to_batch: a dictionary of properties to be batched.
            constants across batch are defined in the generate_variables function
        generate_variables: function that takes batched properties and return
            a graphQL payload for request for this batch
        request: the GraphQL request to call,
        batch_size: the size of the batches to produce
        last_batch_callback: a function that takes the last batch and the result of
            this method as arguments

    Example:
        ```python
        properties_to_batch={prop1: [0,1], prop2: ['a', 'b']}
        def generate_variables(batched_properties):
            return {
                graphQL_prop1: batched_properties['prop1']
                graphQL_prop2: batched_properties['prop2']
            }
        mutate_from_paginated_call(
                properties_to_batch=properties_to_batch,
                generate_variables=generate_variables
                request= GQL_APPEND_MANY_ASSETS
        )
        ```
    """
    results = []
    batch = None
    for batch_number, batch in enumerate(batch_object_builder(properties_to_batch, batch_size)):
        payload = generate_variables(batch)
        try:
            result = kili.graphql_client.execute(request, payload)
        except GraphQLError as err:
            raise GraphQLError(error=err.error, batch_number=batch_number) from err
        results.append(result)

    sleep(1)  # wait for the backend to process the mutations
    if batch and results and last_batch_callback:
        last_batch_callback(batch, results)
    return results


T = TypeVar("T")


def batcher(iterable: Iterable[T], batch_size: int) -> Generator[List[T], None, None]:
    """Break iterable into sub-iterables with batch_size elements each.

    The last yielded list will have fewer than n elements if the
    length of iterable is not divisible by batch_size:
    """
    iterator = iter(iterable)
    while batch := list(islice(iterator, batch_size)):
        yield batch



"""Common and generic functions to import files into a project."""

import abc
import logging
import mimetypes
import os
import warnings
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from json import dumps, loads
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)
from uuid import uuid4

from tenacity import Retrying, stop_after_attempt, stop_never
from tenacity.retry import retry_if_exception_type
from tenacity.wait import wait_exponential

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.graphql.operations.asset.mutations import (
    GQL_APPEND_MANY_ASSETS,
    GQL_APPEND_MANY_FRAMES_TO_DATASET,
)
from kili.core.helpers import RetryLongWaitWarner, T, format_result, get_mime_type, is_url
from kili.core.utils.pagination import batcher
from kili.domain.asset import AssetFilters
from kili.domain.organization import OrganizationFilters
from kili.domain.project import InputType, ProjectId
from kili.domain.types import ListOrTuple
from kili.services.asset_import.constants import (
    IMPORT_BATCH_SIZE,
    project_compatible_mimetypes,
)
from kili.services.asset_import.exceptions import (
    BatchImportError,
    ImportValidationError,
    MimeTypeError,
    UploadFromLocalDataForbiddenError,
)
from kili.services.asset_import.types import AssetLike, KiliResolverAsset
from kili.utils import bucket
from kili.utils.tqdm import tqdm

if TYPE_CHECKING:
    from kili.client import Kili


class BatchParams(NamedTuple):
    """Contains all parameters related to the batch to import."""

    is_asynchronous: bool
    is_hosted: bool


class ProcessingParams(NamedTuple):
    """Contains all parameters related to the assets processing."""

    raise_error: bool
    verify: bool


class ProjectParams(NamedTuple):
    """Contains all parameters related to the batch to import."""

    project_id: ProjectId
    input_type: InputType


class LoggerParams(NamedTuple):
    """Contains all parameters related to logging."""

    disable_tqdm: Optional[bool]


class BaseBatchImporter:  # pylint: disable=too-many-instance-attributes
    """Base class for BatchImporters."""

    def __init__(
        self, kili: "Kili", project_params: ProjectParams, batch_params: BatchParams, pbar: tqdm
    ) -> None:
        self.kili = kili
        self.project_id = project_params.project_id
        self.input_type = project_params.input_type
        self.is_hosted = batch_params.is_hosted
        self.is_asynchronous = batch_params.is_asynchronous
        self.pbar = pbar
        self.http_client = kili.http_client

        logging.basicConfig()
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

    def import_batch(self, assets: ListOrTuple[AssetLike], verify: bool) -> List[str]:
        """Base actions to import a batch of asset.

        Returns:
            created_assets_ids: list of ids of the created assets
        """
        assets = self.loop_on_batch(self.stringify_metadata)(assets)
        assets = self.loop_on_batch(self.stringify_json_content)(assets)
        assets_ = self.loop_on_batch(self.fill_empty_fields)(assets)
        created_assets_ids = self.import_to_kili(assets_)
        if verify:
            self.verify_batch_imported(assets)
        self.pbar.update(n=len(assets))
        return created_assets_ids

    def verify_batch_imported(self, assets: List):
        """Verifies that the batch has been imported successfully."""
        if self.is_asynchronous:
            logger_func = self.logger.info
            log_message = (
                "Import of assets is taking a long time to complete. This maybe be due to files"
                " being processed by the server."
            )
        else:
            logger_func = self.logger.warning
            log_message = (
                "Import of assets is taking a long time to complete. This may be due to a large"
                " number of assets to be processed by the server."
            )
        for attempt in Retrying(
            # adds at least a limit for asynch imports to the number of time to wait = ~ 5 minutes
            stop=stop_after_attempt(40) if self.is_asynchronous else stop_never,
            retry=retry_if_exception_type(BatchImportError),
            wait=wait_exponential(multiplier=1, min=1, max=8),
            before_sleep=RetryLongWaitWarner(logger_func=logger_func, warn_message=log_message),
            reraise=True,
        ):
            with attempt:
                assets_ids = [assets[-1]["id"]]  # check last asset of the batch only
                filters = AssetFilters(project_id=self.project_id, asset_id_in=assets_ids)
                nb_assets_in_kili = self.kili.kili_api_gateway.count_assets(filters)
                if len(assets_ids) != nb_assets_in_kili:
                    raise BatchImportError(
                        "Number of assets to upload is not equal to number of assets uploaded in"
                        " Kili."
                    )

    def add_ids(self, assets: List[AssetLike]):
        """Adds ids to all assets."""
        return self.loop_on_batch(self.add_id_to_asset)(assets)

    def generate_project_bucket_path(self) -> str:
        """Returns ids."""
        return f"projects/{self.project_id}/assets"

    @staticmethod
    def stringify_metadata(asset: AssetLike) -> AssetLike:
        """Stringify the metadata."""
        json_metadata = asset.get("json_metadata", {})
        if not isinstance(json_metadata, str):
            json_metadata = dumps(json_metadata)
        return {**asset, "json_metadata": json_metadata}

    @staticmethod
    def stringify_json_content(asset: AssetLike) -> AssetLike:
        """Stringify the metadata."""
        json_content = asset.get("json_content", "")
        if not isinstance(json_content, str):
            json_content = dumps(json_content)
        return {**asset, "json_content": json_content}

    @staticmethod
    def add_id_to_asset(asset: AssetLike) -> AssetLike:
        """Generates an asset id."""
        return {**asset, "id": asset.get("id", bucket.generate_unique_id())}

    @staticmethod
    def fill_empty_fields(asset: AssetLike):
        """Fill empty fields with their default value."""
        asset_fields_default_value = AssetLike(
            content="",
            multi_layer_content=None,
            json_content="",
            external_id=uuid4().hex,
            json_metadata="{}",
            is_honeypot=False,
            id="",
        )
        field_names = asset_fields_default_value.keys()
        return KiliResolverAsset(
            **{field: asset.get(field, asset_fields_default_value[field]) for field in field_names}
        )

    @staticmethod
    def loop_on_batch(
        func: Callable[[AssetLike], T],
    ) -> Callable[[ListOrTuple[AssetLike]], List[T]]:
        """Apply a function, that takes a single asset as input, on the whole batch."""

        def loop_func(assets: ListOrTuple[AssetLike]):
            return [func(asset) for asset in assets]

        return loop_func

    @staticmethod
    def build_url_from_parts(*parts) -> str:
        """Builds an url from the parts."""
        return "/".join(parts)

    @staticmethod
    def are_native_videos(assets) -> bool:
        """Determine if assets should be imported asynchronously and cut into frames."""
        should_use_native_video_array = []
        for asset in assets:
            json_metadata = asset.get("json_metadata", "{}")
            json_metadata_ = loads(json_metadata)
            processing_parameters = json_metadata_.get("processingParameters", {})
            should_use_native_video_array.append(
                processing_parameters.get("shouldUseNativeVideo", True)
            )
        if all(should_use_native_video_array):
            return True
        if all(not b for b in should_use_native_video_array):
            return False
        raise ImportValidationError(
            """
            Cannot upload videos to split into frames
            and video to keep as native in the same time.
            Please separate the assets into 2 calls
            """
        )

    def _async_import_to_kili(self, assets: List[KiliResolverAsset]):
        """Import assets with asynchronous resolver."""
        if self.input_type == "IMAGE":
            upload_type = "GEO_SATELLITE"
        elif self.input_type in ("VIDEO", "VIDEO_LEGACY"):
            upload_type = "NATIVE_VIDEO" if self.are_native_videos(assets) else "FRAME_VIDEO"
        else:
            raise NotImplementedError(
                f"Import of {self.input_type} assets is not supported in async calls"
            )

        content = (
            {
                "multiLayerContentArray": [asset["multi_layer_content"] for asset in assets],
                "jsonContentArray": [asset["json_content"] for asset in assets],
            }
            if all(asset.get("multi_layer_content") is not None for asset in assets)
            else {"contentArray": [asset["content"] for asset in assets]}
        )
        payload = {
            "data": {
                **content,
                "externalIDArray": [asset["external_id"] for asset in assets],
                "idArray": [asset["id"] for asset in assets],
                "jsonMetadataArray": [asset["json_metadata"] for asset in assets],
                "uploadType": upload_type,
            },
            "where": {"id": self.project_id},
        }
        result = self.kili.graphql_client.execute(GQL_APPEND_MANY_FRAMES_TO_DATASET, payload)
        format_result("data", result, None, self.kili.http_client)
        return []

    def _sync_import_to_kili(self, assets: List[KiliResolverAsset]):
        """Import assets with synchronous resolver."""
        payload = {
            "data": {
                "contentArray": [asset["content"] for asset in assets],
                "externalIDArray": [asset["external_id"] for asset in assets],
                "idArray": [asset["id"] for asset in assets],
                "isHoneypotArray": [asset["is_honeypot"] for asset in assets],
                "jsonContentArray": [asset["json_content"] for asset in assets],
                "jsonMetadataArray": [asset["json_metadata"] for asset in assets],
            },
            "where": {"id": self.project_id},
        }
        result = self.kili.graphql_client.execute(GQL_APPEND_MANY_ASSETS, payload)
        created_assets = format_result("data", result, None, self.kili.http_client)
        return [asset["id"] for asset in created_assets]

    def import_to_kili(self, assets: List[KiliResolverAsset]):
        """Import assets to Kili with the right resolver.

        Returns:
            created_assets_ids: list of ids of the created assets
        """
        if self.is_asynchronous:
            return self._async_import_to_kili(assets)
        return self._sync_import_to_kili(assets)


class ContentBatchImporter(BaseBatchImporter):
    """Class defining the methods to import a batch of assets with content."""

    def import_batch(self, assets: List[AssetLike], verify: bool):
        """Method to import a batch of asset with content."""
        assets = self.add_ids(assets)
        if not self.is_hosted:
            assets_with_content = [
                asset
                for asset in assets
                if asset.get("content") or asset.get("multi_layer_content")
            ]
            assets = [
                asset
                for asset in assets
                if not asset.get("content") and not asset.get("multi_layer_content")
            ]
            if len(assets_with_content) > 0:
                assets += self.upload_local_content_to_bucket(assets_with_content)
        return super().import_batch(assets, verify)

    def get_content_type_and_data_from_content(
        self, content: Optional[Union[str, bytes]]
    ) -> Tuple[bytes, Optional[str]]:
        """Returns the data of the content (path) and its content type."""
        assert content
        assert isinstance(content, str)
        with Path(content).open("rb") as file:
            data = file.read()
            content_type, _ = mimetypes.guess_type(content)
            return data, content_type

    def get_type_and_data_from_content_array(
        self, content_array: List[Optional[Union[str, bytes]]]
    ) -> List[Tuple[Union[bytes, str], Optional[str]]]:
        # pylint:disable=line-too-long
        """Returns the data of the content (path) and its content type for each element in the array."""
        return list(map(self.get_content_type_and_data_from_content, content_array))

    def upload_local_content_to_bucket(self, assets: List[AssetLike]):
        """Upload local content to a bucket."""
        project_bucket_path = self.generate_project_bucket_path()
        # tuple containing (bucket_path, file_path, asset_index, content_index)
        to_upload: List[Tuple[str, Any, int, Union[int, None]]] = []
        for i, asset in enumerate(assets):
            asset_id = asset.get("id", bucket.generate_unique_id())
            multi_layer_content = asset.get("multi_layer_content")
            if multi_layer_content:
                for j, item in enumerate(multi_layer_content):
                    bucket_path = BaseBatchImporter.build_url_from_parts(
                        project_bucket_path, asset_id, "content", str(j)
                    )
                    to_upload.append((bucket_path, item.get("path"), i, j))
            else:
                bucket_path = BaseBatchImporter.build_url_from_parts(
                    project_bucket_path, asset_id, "content"
                )
                to_upload.append((bucket_path, asset.get("content"), i, None))
        signed_urls = bucket.request_signed_urls(
            self.kili, [bucket_path for bucket_path, *_ in to_upload]
        )
        data_and_content_type_array = self.get_type_and_data_from_content_array(
            [file_path for _, file_path, *_ in to_upload]
        )
        data_array, content_type_array = zip(*data_and_content_type_array)
        with ThreadPoolExecutor() as threads:
            url_gen = threads.map(
                bucket.upload_data_via_rest,
                signed_urls,
                data_array,
                content_type_array,
                repeat(self.http_client),
            )
        assets_with_content = []
        for asset in assets:
            asset_copy = asset.copy()
            multi_layer_content = asset.get("multi_layer_content")
            if multi_layer_content:
                asset_copy["multi_layer_content"] = [
                    {key: value for key, value in content.items() if key != "path"}
                    for content in multi_layer_content
                ]
            assets_with_content.append(asset_copy)
        for (_, _, asset_index, content_index), url in zip(to_upload, url_gen):
            if content_index is not None:
                assets_with_content[asset_index]["multi_layer_content"][content_index]["url"] = url
            else:
                assets_with_content[asset_index]["content"] = url
        return assets_with_content


class JsonContentBatchImporter(BaseBatchImporter):
    """Class defining the import methods for a batch of assets twith json_content."""

    @staticmethod
    def stringify_json_content(asset: AssetLike):
        """Stringify the json content if not a str."""
        json_content = asset.get("json_content", {})
        if not isinstance(json_content, str):
            json_content = dumps(json_content)
        return AssetLike(**{**asset, "json_content": json_content})  # type: ignore

    def upload_json_content_to_bucket(self, assets: List[AssetLike]):
        """Upload the json_contents to a bucket with signed urls."""
        project_bucket_path = self.generate_project_bucket_path()
        asset_json_content_paths = [
            BaseBatchImporter.build_url_from_parts(
                project_bucket_path,
                asset.get("id", bucket.generate_unique_id()),
                "jsonContent",
            )
            for asset in assets
        ]
        signed_urls = bucket.request_signed_urls(self.kili, asset_json_content_paths)
        json_content_array = [asset.get("json_content") for asset in assets]
        with ThreadPoolExecutor() as threads:
            url_gen = threads.map(
                bucket.upload_data_via_rest,
                signed_urls,
                json_content_array,
                ["text/plain"] * len(assets),
                repeat(self.http_client),
            )
        # pylint: disable=line-too-long
        return [AssetLike(**{**asset, "json_content": url}) for asset, url in zip(assets, url_gen)]  # type: ignore

    def import_batch(self, assets: List[AssetLike], verify: bool):
        """Method to import a batch of asset with json content."""
        assets = self.add_ids(assets)
        assets = self.loop_on_batch(self.stringify_json_content)(assets)
        assets = self.upload_json_content_to_bucket(assets)
        return super().import_batch(assets, verify)


class BaseAbstractAssetImporter(abc.ABC):
    """Base class for DataImporters classes."""

    def __init__(
        self,
        kili: "Kili",
        project_params: ProjectParams,
        processing_params: ProcessingParams,
        logger_params: LoggerParams,
    ) -> None:
        self.kili = kili
        self.project_params = project_params
        self.raise_error = processing_params.raise_error
        self.verify = processing_params.verify
        self.pbar = tqdm(disable=logger_params.disable_tqdm)

    @abc.abstractmethod
    def import_assets(self, assets: List[AssetLike]) -> List[str]:
        """Import assets into Kili.

        Returns:
            created_assets_ids: list of ids of the created assets
        """
        raise NotImplementedError

    @staticmethod
    def is_hosted_content(assets: List[AssetLike]) -> bool:
        """Determine if the assets to upload are from local files or hosted data.

        Raise an error if a mix of both.
        """
        multi_layer_contents = [
            asset.get("multi_layer_content") for asset in assets if asset.get("multi_layer_content")
        ]
        if len(multi_layer_contents) > 0:
            if any(
                not isinstance(content.get("path"), str)
                for multi_layer_content in multi_layer_contents
                for content in multi_layer_content  # type: ignore
                if content
            ):
                raise ImportValidationError(
                    "Cannot import multi_layer_content with empty path. Please provide a path for"
                    " each multi_layer_content"
                )
            return False
        contents = [asset.get("content") for asset in assets if asset.get("content")]
        if all(is_url(content) for content in contents):
            return True
        if any(is_url(content) for content in contents):
            raise ImportValidationError(
                "Cannot upload hosted data and local files at the same time. Please separate the"
                " assets into 2 calls."
            )
        return False

    @staticmethod
    def check_asset_contents(assets: List[AssetLike]) -> None:
        """Determine if the assets have at least one content or json_content.

        Raise an error if not
        """
        # Raise an error if there is an asset with no content,
        # no multi_layer_content and no json_content
        for asset in assets:
            if (
                not asset.get("content")
                and not asset.get("multi_layer_content")
                and not asset.get("json_content")
            ):
                raise ImportValidationError(
                    "Cannot import asset with empty content, empty"
                    " multi_layer_content and empty json_content"
                )

    def _can_upload_from_local_data(self) -> bool:
        user_me = self.kili.kili_api_gateway.get_current_user(fields=("email",))
        options = QueryOptions(first=1, disable_tqdm=True)
        organization = self._get_organization(user_me["email"], options)
        return organization["license"]["uploadLocalData"]

    def _get_organization(self, email: str, options: QueryOptions) -> Dict:
        return next(
            self.kili.kili_api_gateway.list_organizations(
                filters=OrganizationFilters(email=email),
                fields=["license.uploadLocalData"],
                description="",
                options=options,
            )
        )

    def _check_upload_is_allowed(self, assets: List[AssetLike]) -> None:
        # TODO: avoid querying API for each asset to upload when doing this check
        if not self.is_hosted_content(assets) and not self._can_upload_from_local_data():
            raise UploadFromLocalDataForbiddenError("Cannot upload content from local data")

    def filter_local_assets(self, assets: List[AssetLike], raise_error: bool):
        """Filter out local files that cannot be imported.

        Return an error at the first file that cannot be imported if raise_error is True
        """
        filtered_assets = []
        for asset in assets:
            json_content = asset.get("json_content")
            multi_layer_content = asset.get("multi_layer_content")
            path = asset.get("content")
            if multi_layer_content or (json_content and not path):
                filtered_assets.append(asset)
                continue
            assert path
            assert isinstance(path, str)
            try:
                self.check_mime_type_compatibility(path)
                filtered_assets.append(asset)
            except (FileNotFoundError, MimeTypeError):
                if raise_error:
                    raise
        if len(filtered_assets) == 0:
            # pylint: disable=line-too-long
            raise ImportValidationError(
                """No files to upload. Check that the paths exist and file types are compatible with the project."""
            )
        return filtered_assets

    def check_mime_type_compatibility(self, path: str):
        """Check that the mimetype of a local file is compatible with the project input type.

        Raise:
             FileNotFoundError: if path do not exists
             MimeTypeError: if the asset is not compatible with the project type
        """
        if not os.path.isfile(path):
            raise FileNotFoundError(f"file {path} does not exist")
        mime_type = get_mime_type(path)
        if mime_type is None:
            raise MimeTypeError(f"The mime type of the asset {path} has not been found")

        input_type = self.project_params.input_type
        if mime_type not in project_compatible_mimetypes[input_type]:
            raise MimeTypeError(
                f"File mime type for {path} is {mime_type} and does not correspond to the type of"
                " the project. File mime type should be one of"
                f" {project_compatible_mimetypes[input_type]}"
            )
        return True

    def filter_duplicate_external_ids(self, assets):
        """Filter out assets whose external_id is already in the project."""
        if len(assets) == 0:
            raise ImportValidationError("No assets to import")
        assets_in_project = self.kili.kili_api_gateway.list_assets(
            AssetFilters(project_id=self.project_params.project_id),
            ["externalId"],
            QueryOptions(disable_tqdm=True),
        )
        external_ids_in_project = [asset["externalId"] for asset in assets_in_project]
        filtered_assets = [
            asset for asset in assets if asset.get("external_id") not in external_ids_in_project
        ]
        if len(filtered_assets) == 0:
            raise ImportValidationError(
                "No assets to import, all given external_ids already exist in the project"
            )
        nb_duplicate_assets = len(assets) - len(filtered_assets)
        if nb_duplicate_assets > 0:
            warnings.warn(
                f"{nb_duplicate_assets} assets were not imported because their external_id are"
                " already in the project",
                stacklevel=2,
            )
        return filtered_assets

    def import_assets_by_batch(
        self,
        assets: List[AssetLike],
        batch_importer: BaseBatchImporter,
        batch_size=IMPORT_BATCH_SIZE,
    ):
        """Split assets by batch and import them with a given batch importer."""
        batch_generator = batcher(assets, batch_size)
        nb_batch = (len(assets) - 1) // batch_size + 1
        self.pbar.total = len(assets)
        self.pbar.refresh()

        created_asset_ids: List[str] = []
        for i, batch_assets in enumerate(batch_generator):
            # check last batch only
            verify = i == (nb_batch - 1) and self.verify
            created_asset_ids += batch_importer.import_batch(batch_assets, verify)
        return created_asset_ids



"""Queries of project mutations."""

from .fragments import PROJECT_FRAGMENT, PROJECT_FRAGMENT_ID, ROLE_FRAGMENT

GQL_APPEND_TO_ROLES = f"""
mutation($data: AppendToRolesData!, $where: ProjectWhere!) {{
  data: appendToRoles(
    data: $data
    where: $where
  ) {{
      {PROJECT_FRAGMENT}
  }}
}}
"""

GQL_UPDATE_PROPERTIES_IN_PROJECT = f"""
mutation(
    $canNavigateBetweenAssets: Boolean
    $canSkipAsset: Boolean
    $consensusMark: Float
    $consensusTotCoverage: Int
    $description: String
    $honeypotMark: Float
    $instructions: String
    $inputType: InputType
    $jsonInterface: String
    $metadataTypes: JSON
    $minConsensusSize: Int
    $numberOfAssets: Int
    $numberOfSkippedAssets: Int
    $numberOfRemainingAssets: Int
    $numberOfReviewedAssets: Int
    $projectID: ID!
    $reviewCoverage: Int
    $title: String
    $useHoneyPot: Boolean
    $archived: Boolean
) {{
  data: updatePropertiesInProject(
    where: {{
      id: $projectID
    }},
    data: {{
      canNavigateBetweenAssets: $canNavigateBetweenAssets
      canSkipAsset: $canSkipAsset
      consensusMark: $consensusMark
      consensusTotCoverage: $consensusTotCoverage
      description: $description
      honeypotMark: $honeypotMark
      instructions: $instructions
      inputType: $inputType
      jsonInterface: $jsonInterface
      metadataTypes: $metadataTypes
      minConsensusSize: $minConsensusSize
      numberOfAssets: $numberOfAssets
      numberOfSkippedAssets: $numberOfSkippedAssets
      numberOfRemainingAssets: $numberOfRemainingAssets
      numberOfReviewedAssets: $numberOfReviewedAssets
      reviewCoverage: $reviewCoverage
      title: $title
      useHoneyPot: $useHoneyPot
      archived: $archived
    }}
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""


GQL_UPDATE_PROPERTIES_IN_ROLE = f"""
mutation(
    $roleID: ID!
    $projectID: ID!
    $userID: ID!
    $role: ProjectRole!
) {{
  data: updatePropertiesInRole(
    where: {{id: $roleID}}
    data: {{
      projectID: $projectID
      userID: $userID
      role: $role
    }}) {{
      {ROLE_FRAGMENT}
  }}
}}
"""

GQL_DELETE_FROM_ROLES = f"""
mutation($where: ProjectUserWhere!) {{
  data: deleteFromRoles(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""


GQL_DELETE_PROJECT = f"""
mutation($projectID: ID!) {{
  data: deleteProject(where: {{
      id: $projectID
    }}
    ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_PROJECT_DELETE_ASYNCHRONOUSLY = """
mutation($where: ProjectWhere!) {
  data: deleteProjectAsynchronously(where: $where)
}
"""

GQL_PROJECT_UPDATE_ANONYMIZATION = f"""
mutation($input: UpdateProjectAnonymizationInput!) {{
  data: updateProjectAnonymization(input: $input) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""



"""GraphQL module."""

from abc import ABC, abstractmethod
from typing import Callable, Dict, Generator, Optional, Type, TypeVar

from kili.adapters.http_client import HttpClient
from kili.adapters.kili_api_gateway.helpers.queries import (
    QueryOptions,
    fragment_builder,
)
from kili.core.constants import QUERY_BATCH_SIZE
from kili.core.helpers import format_result
from kili.domain.types import ListOrTuple
from kili.utils.tqdm import tqdm

from .graphql_client import GraphQLClient

T = TypeVar("T")


class BaseQueryWhere(ABC):
    """Abtsract class for defining the where payload to send in a graphQL query."""

    def __init__(self) -> None:
        self._graphql_payload = self.graphql_where_builder()

    @abstractmethod
    def graphql_where_builder(self) -> Dict:
        """Build the GraphQL where payload from the arguments given to the where class."""
        raise NotImplementedError

    @property
    def graphql_payload(self) -> Dict:
        """Where payload to send in the graphQL query."""
        return self._graphql_payload


class GraphQLQuery(ABC):
    """Query class for querying Kili objects.

    It factorizes code for executing paginated queries
    """

    def __init__(self, client: GraphQLClient, http_client: HttpClient) -> None:
        self.client = client
        self.http_client = http_client

    @staticmethod
    @abstractmethod
    def query(fragment: str) -> str:
        """Return the GraphQL object query to be sent, according to the fragment to return.

        Args:
            fragment: the fragment to return in the query

        Return:
            The query to be sent to graphQL
        """
        raise NotImplementedError

    COUNT_QUERY: str = NotImplemented
    FORMAT_TYPE: Optional[Type] = None

    def __call__(
        self,
        where: BaseQueryWhere,
        fields: ListOrTuple[str],
        options: QueryOptions,
        post_call_function: Optional[Callable] = None,
    ) -> Generator[Dict, None, None]:
        # pylint: disable=line-too-long
        """Get a generator of objects of the specified type in accordance with the provided where."""
        fragment = fragment_builder(fields)
        query = self.query(fragment)

        return self.execute_query_from_paginated_call(query, where, options, post_call_function)

    def count(self, where: BaseQueryWhere) -> int:
        """Count the number of objects matching the given where payload."""
        payload = {"where": where.graphql_payload}
        count_result = self.client.execute(self.COUNT_QUERY, payload)
        return self.format_result("data", count_result, int)

    def get_number_of_elements_to_query(self, where: BaseQueryWhere, options: QueryOptions) -> int:
        """Return the total number of element to query for one query.

        It uses both the argument first given by the user
        and the total number of available objects obtained with a graphQL count query
        """
        first = options.first
        skip = options.skip
        count_rows_available = self.count(where)
        count_objects_queried = max(count_rows_available - skip, 0)
        if first is None:
            return count_objects_queried
        return min(count_objects_queried, first)

    def execute_query_from_paginated_call(
        self,
        query: str,
        where: BaseQueryWhere,
        options: QueryOptions,
        post_call_function: Optional[Callable],
    ) -> Generator[Dict, None, None]:
        """Build a row generator from paginated calls.

        Args:
            query: The object query to execute and to send to graphQL, in string format
            where: The where payload to be sent to graphQL,
                as a value of the 'where' key in the global payload
            options: The query options
            post_call_function: A function to be applied to the result of the query
        """
        # we can get the total number of elements to query
        if isinstance(self.COUNT_QUERY, str):
            nb_rows_to_query = self.get_number_of_elements_to_query(where, options)
            disable_tqdm = options.disable_tqdm

        # we don't have count methods but we know the total number of elements to query
        elif options.first is not None:
            nb_rows_to_query = options.first
            disable_tqdm = options.disable_tqdm

        # we don't have count methods and we don't know the total number of elements to query
        else:
            nb_rows_to_query = None
            disable_tqdm = True

        if nb_rows_to_query == 0:
            yield from ()
        else:
            with tqdm(total=nb_rows_to_query, disable=disable_tqdm) as pbar:
                count_rows_retrieved = 0
                while True:
                    if nb_rows_to_query is not None and count_rows_retrieved >= nb_rows_to_query:
                        break

                    skip = count_rows_retrieved + options.skip
                    first = (
                        min(QUERY_BATCH_SIZE, nb_rows_to_query - count_rows_retrieved)
                        if nb_rows_to_query is not None
                        else QUERY_BATCH_SIZE
                    )
                    payload = {"where": where.graphql_payload, "skip": skip, "first": first}
                    rows = self.client.execute(query, payload)
                    rows = self.format_result("data", rows, self.FORMAT_TYPE)

                    if rows is None or len(rows) == 0:
                        break

                    if post_call_function is not None:
                        rows = post_call_function(rows)

                    if isinstance(rows, Dict):
                        yield rows
                        break

                    yield from rows

                    count_rows_retrieved += len(rows)
                    pbar.update(len(rows))

                    if len(rows) < first:
                        break

    def format_result(self, name: str, result: dict, object_: Optional[Type[T]] = None) -> T:
        """Format the result of a graphQL query."""
        return format_result(name, result, object_, self.http_client)



"""Organization domain."""

from dataclasses import dataclass
from datetime import datetime
from typing import NewType, Optional

OrganizationId = NewType("OrganizationId", str)


@dataclass(frozen=True)
class OrganizationFilters:
    """Dataclass to the OrganizationQuery to filter the results."""

    email: Optional[str] = None
    organization_id: Optional[OrganizationId] = None


@dataclass(frozen=True)
class OrganizationMetricsFilters:
    """Dataclass to the OrganizationQuery to filter the results."""

    id: OrganizationId
    start_datetime: datetime
    end_datetime: datetime


@dataclass(frozen=True)
class OrganizationToCreateInput:
    """Organization to create use case input."""

    name: str
    address: str
    city: str
    country: str
    zip_code: str


@dataclass(frozen=True)
class OrganizationToUpdateInput:
    """Organization to update use case input."""

    name: Optional[str] = None
    license: Optional[dict] = None



"""Notification domain."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, NewType, Optional

if TYPE_CHECKING:
    from .user import UserFilter

NotificationId = NewType("NotificationId", str)


@dataclass
class NotificationFilter:
    """Notification filter."""

    has_been_seen: Optional[bool]
    id: Optional[NotificationId]
    user: Optional["UserFilter"]



"""CLI's project member list subcommand."""

from typing import Optional

import click
import pandas as pd
from tabulate import tabulate

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.graphql.operations.project_user.queries import (
    ProjectUserQuery,
    ProjectUserWhere,
)
from kili.entrypoints.cli.common_args import Arguments, Options
from kili.entrypoints.cli.helpers import get_kili_client

ROLE_ORDER = {v: i for i, v in enumerate(["ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"])}


@click.command(name="list")
@Options.api_key
@Options.endpoint
@Arguments.project_id
@Options.tablefmt
def list_members(api_key: Optional[str], endpoint: Optional[str], project_id: str, tablefmt: str):
    """List the members of the project.

    \b
    !!! Examples
        ```
        kili project member list <project_id> --stdout-format pretty
        ```
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)
    members_list = list(
        ProjectUserQuery(kili.graphql_client, kili.http_client)(
            where=ProjectUserWhere(project_id=project_id),
            fields=[
                "role",
                "activated",
                "user.email",
                "user.id",
                "user.firstname",
                "user.lastname",
                "user.organization.name",
            ],
            options=QueryOptions(disable_tqdm=True),
        )
    )
    members = pd.DataFrame(members_list)
    # pylint: disable=line-too-long
    members = pd.concat([members.drop(["user"], axis=1), members["user"].apply(pd.Series)], axis=1)  # type: ignore
    members = pd.concat(
        [members.drop(["organization"], axis=1), members["organization"].apply(pd.Series)],
        axis=1,
    )
    members = members.loc[members["activated"]]
    members = members.rename(
        columns={"role": "ROLE", "email": "EMAIL", "id": "ID", "name": "ORGANIZATION"},
    )
    members["NAME"] = members["firstname"].str.title() + " " + members["lastname"].str.title()
    members = members.sort_values(
        by=["ROLE", "lastname"],
        ascending=True,
        key=lambda column: column.map(ROLE_ORDER) if column.name == "ROLE" else column,
    )
    members = members[["ROLE", "NAME", "EMAIL", "ID", "ORGANIZATION"]]
    print(tabulate(members, headers="keys", tablefmt=tablefmt, showindex=False))  # type: ignore



"""Question use cases."""

from dataclasses import dataclass
from typing import List, Optional, cast

from kili.adapters.kili_api_gateway.issue.types import IssueToCreateKiliAPIGatewayInput
from kili.domain.asset import AssetExternalId, AssetId
from kili.domain.project import ProjectId
from kili.domain.question import QuestionId
from kili.use_cases.asset.utils import AssetUseCasesUtils
from kili.use_cases.base import BaseUseCases


@dataclass
class QuestionToCreateUseCaseInput:
    """Question to create use case input."""

    text: Optional[str]
    asset_id: Optional[AssetId]
    asset_external_id: Optional[AssetExternalId]


class QuestionUseCases(BaseUseCases):
    """Question use cases."""

    def create_questions(
        self,
        project_id: ProjectId,
        questions: List[QuestionToCreateUseCaseInput],
    ) -> List[QuestionId]:
        """Create questions."""
        if questions[0].asset_id is not None:
            # we assume that if 1 question asset Id is not None, all there others are too
            asset_id_array = [cast(AssetId, question.asset_id) for question in questions]
            external_id_array = None
        else:
            asset_id_array = None
            external_id_array = [
                cast(AssetExternalId, question.asset_external_id) for question in questions
            ]

        asset_ids = AssetUseCasesUtils(self._kili_api_gateway).get_asset_ids_or_throw_error(
            asset_ids=asset_id_array, external_ids=external_id_array, project_id=project_id
        )

        gateway_issues = [
            IssueToCreateKiliAPIGatewayInput(
                asset_id=asset_id,
                text=question.text,
                object_mid=None,
                label_id=None,
            )
            for (asset_id, question) in zip(asset_ids, questions)
        ]
        return [
            QuestionId(str(id_))
            for id_ in self._kili_api_gateway.create_issues(
                type_="QUESTION", issues=gateway_issues, description="Creating questions"
            )
        ]



"""User use cases."""

from typing import Dict, Generator, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.user.types import (
    CreateUserDataKiliGatewayInput,
    UserDataKiliGatewayInput,
)
from kili.core.enums import OrganizationRole
from kili.domain.organization import OrganizationId
from kili.domain.types import ListOrTuple
from kili.domain.user import UserFilter
from kili.use_cases.base import BaseUseCases


class UserUseCases(BaseUseCases):
    """User use cases."""

    def list_users(
        self, filters: UserFilter, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """List all users."""
        return self._kili_api_gateway.list_users(
            fields=fields, user_filters=filters, options=options
        )

    def count_users(self, filters: UserFilter) -> int:
        """Count users."""
        return self._kili_api_gateway.count_users(user_filters=filters)

    def create_user(
        self,
        email: str,
        password: str,
        organization_role: OrganizationRole,
        firstname: Optional[str],
        lastname: Optional[str],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Create a user."""
        return self._kili_api_gateway.create_user(
            data=CreateUserDataKiliGatewayInput(
                email=email,
                password=password,
                organization_role=organization_role,
                firstname=firstname,
                lastname=lastname,
            ),
            fields=fields,
        )

    def update_password(
        self,
        old_password: str,
        new_password_1: str,
        new_password_2: str,
        user_filter: UserFilter,
        fields: ListOrTuple[str],
    ) -> Dict:
        """Update user password."""
        return self._kili_api_gateway.update_password(
            old_password=old_password,
            new_password_1=new_password_1,
            new_password_2=new_password_2,
            user_filter=user_filter,
            fields=fields,
        )

    def update_user(
        self,
        user_filter: UserFilter,
        firstname: Optional[str],
        lastname: Optional[str],
        organization_id: Optional[OrganizationId],
        organization_role: Optional[OrganizationRole],
        activated: Optional[bool],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Update user."""
        return self._kili_api_gateway.update_user(
            user_filter=user_filter,
            data=UserDataKiliGatewayInput(
                activated=activated,
                firstname=firstname,
                lastname=lastname,
                organization_id=organization_id,
                organization_role=organization_role,
                api_key=None,
                email=None,
                has_completed_labeling_tour=None,
                hubspot_subscription_status=None,
                organization=None,
            ),
            fields=fields,
        )
